<?
/*
   +----------------------------------------------------------------------+
   | Zend installation class                                              |
   +----------------------------------------------------------------------+
   | Copyright (c) 1998-2004 Zend Technologies Ltd.                       |
   +----------------------------------------------------------------------+
   | The contents of this source file is the sole property of             |
   | Zend Technologies Ltd.  Unauthorized duplication or access is        |
   | prohibited.                                                          |
   +----------------------------------------------------------------------+
   | Authors: Michael Spector <michael@zend.com>                          |
   |          Anya Tarnyavsky <anya@zend.com>                             |
   +----------------------------------------------------------------------+
   | Tips:                                                                |
   | 1) If you want to find some function - search it with uppercase      |
   |                                                                      |
   +----------------------------------------------------------------------+
*/

if(defined('DBUILDER')){
	include_once("dbuilder.inc");
}

include_once("gaugebox.inc");

if(defined('SERIAL_NUMBER')){
	include_once("forms.inc");
}

class Install
{
	var $forms;  /* forms for displaying more complex dialogs */

	function Install($product, $version)
	{
		set_time_limit(0);
		error_reporting(E_ALL ^ E_NOTICE);

		print ("Please, wait ...\n");

		$this->conf['product'] = $product;
		$this->conf['version'] = $version;
		$this->conf['title'] = "$product $version";
		$this->conf['backtitle'] = "$product Installation Script";
		$this->conf['logfile_name'] = "install.log";
		$this->conf['packaging_dot_zend'] = "packaging@zend.com";
		$this->conf['backup_suffix'] = "-". strtolower(preg_replace("/[\s\/]+/", "_", $product)).".bak";
		$this->conf['progress_bar_started'] = false;
		$this->conf['wrap_size'] = 75;
		$this->conf['php_ini_opened'] = false;

		$this->conf['logfile'] = fopen($this->conf['logfile_name'], "w");
		if(empty($this->conf['logfile'])){
			$this->on_error("Unable to open install.log for writting.");
		}

		$this->install_log("Starting installation of $product $version ...");

		// define OS
		$this->conf['os'] = strstr(PHP_OS, 'WIN') ? "WINDOWS" : "UNIX";
		if(empty($this->conf['os'])){
			$this->on_error("Undefined operating system: ". PHP_OS);
		}

		$this->install_log("Operating system: ".$this->conf['os']);

		// define system information
		// $uname = posix_uname();
		$this->conf['uname'] = $this->my_uname();
		if(empty($this->conf['uname']['nodename'])){
			$this->conf['uname']['nodename'] = "localhost";
		}

		$this->install_log("System name: ".$this->conf['uname']['sysname'].
				" ".$this->conf['uname']['release'].
				" ".$this->conf['uname']['machine']);

		$this->install_log("Hostname: ".$this->conf['uname']['nodename']);

		if($this->conf['os'] == "WINDOWS"){
			$this->conf['slash'] = "\\";
			$this->conf['nl']  = "\r\n";
			$this->conf['so_ext'] = ".dll";
			$this->conf['zend_ext_directive'] = "zend_extension_ts";
			$this->conf['sysdrive'] = getenv("SYSTEMDRIVE");
			if(empty($this->conf['sysdrive'])){
				$this->conf['sysdrive'] = "C:\\";
			}
		}
		else{ /* if($this->conf['os'] == "UNIX") */
			$this->conf['slash'] = "/";
			$this->conf['nl'] = "\n";
			$this->conf['zend_ext_directive'] = "zend_extension";

			if($this->conf['uname']['sysname'] == "Darwin"){
				$this->conf['so_ext'] = ".so";
			}
			elseif($this->conf['uname']['sysname'] == "HP-UX"){
				$this->conf['so_ext'] = ".sl";
			}
			else{
				$this->conf['so_ext'] = ".so";
			}

			$this->conf['passwdfile'] = "/etc/passwd";
			$this->conf['groupfile'] = "/etc/group";
		}

		// determine temporary directory
		$this->conf['tmp_dir'] = getenv("TMPDIR");
		if(empty($this->conf['tmp_dir'])){
			$this->conf['tmp_dir'] = getenv("TEMP");
			if(empty($this->conf['tmp_dir'])){
				if($this->conf['os'] == "UNIX"){
					$this->conf['tmp_dir'] = "/tmp";
				}
			}
			$this->conf['tmp_dir'] = $this->conf['tmp_dir'].$this->conf['slash']."zendinstall.". getmypid();
		}

		$this->install_log("Setting TEMP=".$this->conf['tmp_dir']);

		// and create it
		$this->my_delete($this->conf['tmp_dir']);
		$this->my_mkdir($this->conf['tmp_dir']);

		$this->conf['interactive'] = true;

		# Read default config values
		$config_file = strtolower(preg_replace("/\s+/", "_", $product)).".conf";
		$dv_file = $this->search_file(array("../", "../data"), array($config_file));
		if(!empty($dv_file)) {
			$dv_file = $this->make_path($dv_file, $config_file);
		}
		if($this->my_file_exists($dv_file)) {
			$default_config = parse_ini_file($dv_file);

			$this->conf['interactive'] = false;
			
			foreach($default_config as $name => $val) {
				$this->conf[$name] = $val;
			}
		}

		// get the basename of the calling script
		$basename = basename(getenv("CALLING_SCRIPT"));

		if($basename == "install.sh" || $basename == "install"){

			if(!$this->conf['interactive']) {
				unset($this->conf['dialog']);
			}
			else {
				if(defined('DBUILDER')){
					$this->conf['dialog'] =& new Dbuilder();
				}
				else{
					$dialog = getcwd()."/dialog";
					if(!$this->my_file_exists($dialog)){
						$this->zend_bug("File: $dialog does not exist!");
					}
					if(!is_executable($dialog)){
						chmod($dialog, 0555);
					}
					$this->conf['dialog'] = $dialog;
				}
			}
		}
		else if($basename == "install-tty"){
			unset($this->conf['dialog']);
		}
		else{
			$this->on_error("Script called under a wrong name: $basename");
		}

		if(defined('SERIAL_NUMBER')){
			$this->forms =& new Forms($this, !empty($this->conf['dialog']));
		}

		if(!empty($this->conf['dialog'])){ // set common dialog setup

			if(defined('DBUILDER')){
				$this->conf['dialog']->set_backtitle($this->conf['backtitle']);
			}
			else{
				$this->conf['dialog'] .= " --clear --backtitle \"".$this->conf['backtitle']."\" --title \"".$this->conf['title']."\"";
				$this->conf['max_x'] = 78; // default size
				$this->conf['max_y'] = 24;
				$this->conf['dialog_tmp'] = $this->conf['tmp_dir'] . "/dialog.tmpfile";
			}

			$this->conf['dialog_error'] = "Error executing dialog utility. ".
				"Most probably this is due to the wrong terminal settings. ".
				"Please use install-tty instead.";
		}

		$this->conf['webserver'] = "webserver";

		$this->conf['comp_dirs_arr'] = array( /* possible directories where to get install components  */
				1 =>  "../data",
				2 =>  "../install",
				3 =>  "../"
				);

		// initialize components array
		$this->conf['components'] = array();

		$this->conf['webserver_docroot_arr'] = array();

		// initialize file permissions
		if($this->my_file_exists("permissions")){
			$file_perms = file("permissions");
			foreach ($file_perms as $num => $line){
				if(preg_match("/^\s*\#/", $line) || preg_match("/^\s*$/", $line)){
					continue;
				}
				list($file, $mode, $owner, $group) = preg_split("/\s+/", $line); // file mode owner group
				$this->conf['file_perms'][$num] = array($file, $mode, $owner, $group);
			}
		}
		else{
			$this->conf['file_perms'] = array();
		}

		$this->conf['build_info'] = array();
		$build_file = $this->search_file(array("../", "../data"), array("_BUILD"));
		if(!empty($build_file)) {
			$build_file = $this->make_path($build_file, "_BUILD");
		}
		if($this->my_file_exists($build_file)) {
			$this->conf['build_info'] = parse_ini_file($build_file);
		}
		else {
			$inventory_file = $this->search_file(array("../", "../data"), array("Inventory.xml"));
			if(!empty($inventory_file)) {
				$inventory_file = $this->make_path($inventory_file, "Inventory.xml");
			}
			if($this->my_file_exists($inventory_file)) {
				$inventory = $this->file2str ($inventory_file);
				if(preg_match ("/\<Package\s+.*version=\"([^\"]+)\".*\>/iU", $inventory, $match)) {
					$this->conf['build_info']['package.version'] = preg_replace("/RC.*$/", "", $match[1]);
				}
				if(preg_match_all ("/\<Component\s([^\>]+)\>/iU", $inventory, $match)) {
					for ($i=0; $i<count($match[0]); $i++) {
						if(preg_match ("/version\=\"([^\"]+)\"/i", $match[1][$i], $ver_match)
							&& preg_match ("/name\=\"([^\"]+)\"/i", $match[1][$i], $name_match)) {
								$this->conf['build_info'][preg_replace("/-php-.*$/", "", $name_match[1]).'.version'] =
									preg_replace("/RC.*$/", "", $ver_match[1]);
						}
					}
				}
			}
			else {
				$this->zend_bug("Can't find file _BUILD or Inventory.xml in the package.");
			}
		}
		
		$this->conf['php_ini_backup_msg'] = "";

		$this->conf['license_agr_file'] = $this->search_file(array("../", "../data"), array("LICENSE"));
		if(!empty($this->conf['license_agr_file'])){
			$this->conf['license_agr_file'] = $this->make_path($this->conf['license_agr_file'], "LICENSE");
		}
		else {
			$this->zend_bug("Can't find file LICENSE in the package.");
		}

		$this->conf['license_host'] = "www.zend.com";
		$this->conf['license_script'] = "licenses/request.php";
		$this->conf['lic_response_file'] = "response.txt";
		$this->conf['lic_error_file'] = "error.txt";
		$this->conf['license_installed'] = false;

		$this->restore_config();
		foreach ($this->conf['saved_config'] as $key => $val) {
			$this->conf[$key] = $val;
		}

		$this->conf['min_space_size'] = $this->my_du("../data");
	}

	// ------------------
	// SYSTEM FUNCTIONS
	// ==================

	// MY_UNAME
	function my_uname()
	{
		$uname = array();
		$uname["sysname"]   = $this->cmd2str("uname -s 2>/dev/null");
		$uname["machine"]   = $this->cmd2str("uname -p 2>/dev/null");
		$uname["release"]   = $this->cmd2str("uname -r 2>/dev/null");
		$uname["nodename"]  = $this->cmd2str("uname -n 2>/dev/null");

		foreach ($uname as $name => $val) {
			if($val == "") {
				$uname[$name] = "unknown";
			}
		}

		return $uname;
	}

	// GUESS_GLIBC_VERSION
	function guess_glibc_version()
	{
		$glibc_ver = "";
		if($this->conf['uname']['sysname'] == "Linux") {
			$glibc_path = $this->cmd2str("ls /lib/libc-?\.?\.* 2>/dev/null");
			if(!empty($glibc_path) && preg_match("/libc-(\d)\.(\d)/", $glibc_path, $match)) {
				$glibc_ver = "$match[1].$match[2]";
			}
		}

		$this->install_log("Detected GLIBC version: $glibc_ver");
		return $glibc_ver;
	}

	// GET_RELEASE_VERSION
	function get_release_version()
	{
		$release = $this->conf['uname']['release'];
		if(preg_match("/^(\d\.\d)/", $this->conf['uname']['release'], $match)) {
			$release = $match[1];
		}

		$this->install_log("Detected release version: $release");
		return $release;
	}

	// CHECK_SYSTEM_SUPPORTED
	function check_system_supported()
	{
		if(!isset($this->conf['supported_systems'][$this->conf['uname']['sysname']])) {
			$error = "OS type: ".$this->conf['uname']['sysname'];
		}
		else {
			$sysconf = $this->conf['supported_systems'][$this->conf['uname']['sysname']];
			if(count($sysconf) > 0) {

				if(isset($sysconf["glibc"])) {
					$glibc_arr = $sysconf["glibc"];
					$glibc_OK = true;
					foreach($glibc_arr as $glibc) {
						$glibc_OK = false;
						if(preg_match("/^([\>\<]?)(.*)$/", $glibc, $match)) {
							$installed_glibc = $this->guess_glibc_version();
							if(  ($match[1] == ">" && $this->ver_cmp($installed_glibc, $match[2]) >= 0)
									|| ($match[1] == "<" && $this->ver_cmp($installed_glibc, $match[2]) <= 0)
									|| $this->ver_equal($installed_glibc, $glibc)) {

								// Glibc version is OK
								$glibc_OK = true;
								break;
							}
						}
					}
					if(!$glibc_OK) {
						$error = "GLIBC version: $installed_glibc";
					}
				}

				if(!isset($error) && isset($sysconf["release"])) {
					$release_arr = $sysconf["release"];
					$release_OK = true;
					foreach($release_arr as $release) {
						$release_OK = false;
						if(preg_match("/^([\>\<]?)(.*)$/", $release, $match)) {
							$installed_release = $this->get_release_version();
							if(  ($match[1] == ">" && $this->ver_cmp($installed_release, $match[2]) >= 0)
									|| ($match[1] == "<" && $this->ver_cmp($installed_release, $match[2]) <= 0)
									|| $this->ver_equal($installed_release, $release)) {

								// Release is OK
								$release_OK = true;
								break;
							}
						}
					}
					if(!$release_OK) {
						$error = "OS release: $installed_release";
					}
				}

				if(!isset($error) && isset($sysconf["machine"])) {
					$machine_arr = $sysconf["machine"];
					$machine_OK = true;
					foreach($machine_arr as $machine) {
						$machine_OK = false;
						if($machine == $this->conf['uname']['machine']) {
							// Processor type is OK
							$machine_OK = true;
						}
					}
					if(!$machine_OK) {
						$error = "processor type: ".$this->conf['uname']['machine'];
					}
				}
			}
		}

		if(isset($error)) {
			$this->yesnobox("The installation detected that this package is not compatible with your system configuration.\n".
					"It seems that you have an unsupported $error\n".
					"Do you still wish to continue the installation?\n".
					"For list of supported configurations, refer to the Zend Product compatibility table.", "NO", true);
		}
	}

	// ------------------
	// BASIC FUNCTIONS
	// ==================

	// CLRSCR
	function clrscr()
	{
		if($this->conf['os'] == "WINDOWS"){
			$this->run_command("cls");
		}
		elseif($this->conf['os'] == "UNIX"){
			$this->run_command("clear");
		}
	}

	// WELCOME_BOX
	function welcome_box($msg="", $addtext="")
	{
		if(!isset($this->conf['dialog'])) { // Clear screen before showing the "welcome" message
			$this->clrscr();
		}

		if($msg == ""){
			$msg = "Welcome to the ".$this->conf['product']." ".$this->conf['version']." Installation Script!\n";
			$msg .= "For more information regarding this script, see the Installation\n";
			$msg .= "section in the ".$this->conf['product']." User Guide.";
			if(!empty($addtext)){
				$msg .= "\n$addtext";
			}
		}
		$this->msgbox($msg);
	}

	// BYE_BYE
	function bye_bye($url="", $msg="", $need_restart=true, $know_port=false)
	{
		$this->clrscr();
		print "\n\n\n\n\n\n\n";
		print "******************************************************************************\n";
		print "\n";
		if(!$this->conf['license_installed']){
			print "  [7;1mYou must [0m[37;44;1mINSTALL the license[0m[7;1m files from ".
				"Zend.com to your installation directory[0m\n";
			print "\n";
		}
		if($need_restart){
			if(isset($this->conf['php_type']) && $this->conf['php_type'] == "fastcgi"){
				print "  [7;1mYou must [0m[37;44;1mRESTART your FastCGI[0m[7;1m for the ".
					"modifications to take effect[0m\n";
				print "\n";
			}
			else{
				print "  [7;1mYou must [0m[37;44;1mRESTART your ".$this->conf['webserver']."[0m[7;1m for the ".
					"modifications to take effect[0m\n";
				print "\n";
			}
		}
		if(!empty($url)) {
			print "  You can access the ".$this->conf['product']."'s User Interface using the URL:\n";
			print "  $url\n";
			print "\n";

			if(!$know_port && !preg_match("@:\d+/@", $url)){
				$url_with_port = preg_replace("@://([^/]*)/@", "://\\1:<port-number>/", $url);

				print "  If your Web server is running on port different from the default\n";
				print "  port 80, you should access the ".$this->conf['product']." as\n";
				print "  $url_with_port\n";
				print "\n";
			}
		}
		if(!empty($msg)){
			print $msg;
		}
		print "\n";
		print "******************************************************************************\n\n\n\n\n\n";

		$this->install_log("Bye bye ... :)");
	}

	// INSTALL_LOG
	function install_log($msg)
	{
		$msg = preg_replace("/[\r\n]+/", " ", $msg);
		$msg = preg_replace("/\s+/", " ", $msg);
		fwrite($this->conf['logfile'], date("[d/m/y H:i:s]")." $msg\n");
		fflush($this->conf['logfile']);
	}

	// SAFE_PCLOSE
	function safe_pclose($fp)
	{
		$status = pclose($fp);
		if(version_compare(PHP_VERSION, "4.3.0", "<") == 1){
			$status = ($status>>8)&0xFF;
		}
		return $status;
	}


	// RUN_COMMAND
	function run_command($cmd, $err_msg="")
	{
		$this->install_log("Executing: $cmd 2>/dev/null");

		$fp = popen("$cmd 2>/dev/null", "w");
		if(empty($fp)){
			$this->on_error("Cannot execute command: $cmd\nERROR: $err_msg");
		}

		$status = $this->safe_pclose($fp);

		$this->install_log("\"$cmd\" returned status: $status");

		return $status;
	}

	// NICE_PATH
	function nice_path($path) /* removes all multiple slashes */
	{
		/* remove multiple slashes from the path */
		$path = preg_replace("/\/\/\/*/", "/", $path);				/* multiple '/' */
		$path = preg_replace("/^\.\//", "", $path);	         /* first './' from the path */
		$path = preg_replace("/\/\.\//", "/", $path);        /* '/./' from the path */
		$path = preg_replace("/^[^\/]+\/\.\.\/?/", "", $path);   /* dir/.. from the beginning */
		$path = preg_replace("/\/[^\/]+\/\.\.\/?/", "/", $path); /* /dir/.. from the path */
		$path = preg_replace("/([^\/])\/+$/", "\\1", $path); /* last / from the end of path */

		return $path;
	}

	// ADD_DASHES
	function add_dashes($msg)
	{
		$dashes_len = 0; /* define dashes line under the menu title */
		$tmp_len = 0;

		for($i=0; $i<strlen($msg); $i++){
			if($msg[$i] == "\n" || $msg[$i] == "\r"){
				if($tmp_len > $dashes_len){
					$dashes_len = $tmp_len;
					$tmp_len = 0;
				}
			}
			else{
				$tmp_len ++;
			}
		}

		if($dashes_len == 0){
			$dashes_len = $tmp_len;
		}

		if($dashes_len > $this->conf['wrap_size']){
			$dashes_len = $this->conf['wrap_size'];
		}

		$dashes = str_repeat("-", $dashes_len);
		return "$msg\n$dashes";
	}

	// MAKE_PATH
	function make_path()
	{
		$args = func_get_args();
		$path = join($this->conf['slash'], $args);
		return ($this->conf['os'] == "WINDOWS")? "\"$path\"" : $path;
	}

	// MAKE_FULL_PATH
	function make_full_path($path, $root)
	{
		if($this->conf['os'] == "UNIX" && !preg_match("/^\//", $path)){
			$path = $this->make_path($root, $path);
		}
		elseif($this->conf['os'] == "WINDOWS" && !preg_match("/^[A-Za-z]\:/", $path)){
			if(preg_match("/^[/\\]/", $path)){
				$path = $this->make_path($this->conf['sysdrive'], $path);
			}
			else{
				$path = $this->make_path($root, $path);
			}
		}

		return $path;
	}

	// MY_DU
	function my_du($filename)
	{
		$total_size = 0;
		if($this->my_dir_exists($filename)) {
			$dir = $this->my_opendir($filename);
			while($file = readdir($dir)){
				if($file == "." || $file == ".."){
					continue;
				}
				$total_size += $this->my_du($this->make_path($filename, $file));
			}
			closedir($dir);
		}
		else if($this->my_file_exists($filename)) {
			$total_size = filesize($filename);
		}

		return $total_size;
	}

	// SIZE_HUMAN
	function size_human($size)
	{
		if($size > 1048576) {
			return round($size/1048576, 1)."MB";
		}
		else if($size > 1024) {
			return round($size/1024, 1)."KB";
		}
		else return round($size, 1)." bytes";
	}

	// EXPAND_TILDE
	function expand_tilde($path) /* converts tilde to a real path */
	{
		if(preg_match("/^~(.*)$/", $path, $match1)) {
			if(preg_match("/^([\w]+)(.*)$/", $match1[1], $match2)) {
				$pwd = $this->run_silently("posix_getpwnam", $match2[1]);
				if(!empty($pwd)) {
					$path = $this->make_path($pwd["dir"], $match2[2]);
				}
			}
			else {
				$path = $this->make_path(getenv("HOME"), $match1[1]);
			}
		}

		return $path;
	}

	// IS_EMAIL_VALID
	function is_email_valid($email)
	{
		return preg_match("/^(.+)\@(.+)\.(.+)$/", $email);
	}

	// IS_IP_VALID
	function is_ip_valid($ip)
	{
		if(preg_match("/^(\d{1,3})\.(\d{1,3})\.(\d{1,3})\.(\d{1,3})$/", $ip, $match)) {
			if($match[1] < 256 && $match[2] < 256 && $match[3] < 256 && $match[4] < 256) {
				return true;
			}
		}
		return false;
	}

	// GET_BUILD_INFO
	function get_build_info()
	{
		return $this->conf['build_info'];
	}

	// INI_ARRAY_REBUILD
	function ini_array_rebuild($ini_arr,$translation_arr) {
		if (!is_array($ini_arr) or !is_array($translation_arr)) return array();
		if (empty($ini_arr) or empty($translation_arr)) return array();
		$result=array();
		foreach ($ini_arr as $directive=>$value) {
			foreach ($translation_arr as $prefix=>$translated_prefix) {
				$prefix_length=strlen($prefix);
				if (strncmp($prefix,$directive,$prefix_length)==0) {
					$translated_directive=$translated_prefix.substr($directive,$prefix_length);
					$result[$translated_directive]=$value;
				}
			}
		}
		return $result;
	}

	// VER_CMP
	function ver_cmp($version1, $version2)
	{
		$vernum1 = preg_replace("/(\d+|x)[\.\_]?(\d+|x)?[\.\_]?(\d+|x)?/", "\\1\\2\\3", $version1);
		$vernum2 = preg_replace("/(\d+|x)[\.\_]?(\d+|x)?[\.\_]?(\d+|x)?/", "\\1\\2\\3", $version2);

		return strcmp($vernum1, $vernum2);
	}

	// VER_EQUAL
	function ver_equal($version1, $version2, $from=0)
	{
		$ver_arr1 = preg_split("/[\.\_]/", $version1);
		$ver_arr2 = preg_split("/[\.\_]/", $version2);

		if(count($ver_arr1) != count($ver_arr2)){
			return false;
		}
		for($i = 0; $i < count($ver_arr1); $i++){
			if(($ver_arr1[$i]!=$ver_arr2[$i]) && ($ver_arr1[$i]!='x') && ($ver_arr2[$i]!='x')){
				return false;
			}
			if(($ver_arr1[$i]!=$ver_arr2[$i]) && ($ver_arr1[$i]=='x') &&
				 ($from != 0) && ($ver_arr2[$i] < $from)){
				return false;
			}
		}

		return true;
	}

	// ARRAY2STRING
	function array2string($array, $index=-1)
	{
		$str = "";
		foreach ($array as $num => $data){
			if($index==-1){
				$str .= $data.$this->conf['nl'];
			}
			else{
				$str .= $data[$index].$this->conf['nl'];
			}
		}
		return $str;
	}

	// ARRAY2PHP_SUPPORT_STRING
	function array2php_support_string($array, $index=-1, $from_index=-1)
	{
		$str = "";
		foreach ($array as $num => $data){

			if($index==-1){
				$str .= $data.$this->conf['nl'];
			}
			else{
				if($from_index != -1 && $data[$from_index] != 0){
					$sub_str=substr($data[$index],0 ,(strlen($data[$index])-1) );
					$str .= $sub_str.$data[$from_index]." and above".$this->conf['nl'];
				}else{
					$str .= $data[$index].$this->conf['nl'];
				}
			}
		}
		return $str;
	}

	// RUN_SILENTLY
	function run_silently($function) /* run function without error outputs */
	{
		$args = func_get_args();
		array_shift($args);
		error_reporting (E_ALL & ~E_WARNING & ~E_NOTICE);

		$func_cmd = '$r = '.$function.' (\'' . join ('\', \'', $args) . '\');';

		eval($func_cmd);

		error_reporting (E_ALL);
		return isset($r) ? $r : null;
	}

	// WARN
	function warn($msg)
	{
		$this->infobox($msg);
	}

	// ON_ERROR
	function on_error($msg = "") /* die with error message */
	{
		$this->install_log("ERROR: $msg");

		if($this->conf['progress_bar_started']) {
			$this->progress_bar_stop();
		}

		if($msg != ""){
			$this->msgbox($this->conf['product']." installation was NOT completed successfully.\n\n$msg\n\n".
					"For further assistance, please contact Zend Support at\nhttp://www.zend.com/support/");
		}
		$this->cleanup();
		die(-1);
	}

	// ON_ABORT
	function on_abort($msg = "", $dont_ask=false) /* die with error message */
	{
		if(!$this->conf['interactive']) {
			$dont_ask = true;
		}

		if(!$dont_ask && !$this->yesnobox("Do you really wish to abort the installation?", "NO")) {
			return;
		}

		$this->install_log("ABORT: $msg");
		$this->save_config();

		if(strlen($msg) > 0) {
			$msg = "\n$msg";
		}
		$this->msgbox($this->conf['product']." installation was aborted.$msg");

		$this->cleanup();
		die(-1);
	}

	// DEFAULT_ABORT
	function default_abort()
	{
		$this->on_abort("The installation was NOT completed successfully.", true);
	}

	// CLEANUP
	function cleanup()
	{
		$this->install_log("Cleaning up ...");

		$this->my_delete($this->conf['tmp_dir']);
		fclose($this->conf['logfile']);
	}

	// SAVE_CONFIG
	function save_config()
	{
		if(!empty($this->conf['saved_config'])) {
			$this->str2file(serialize($this->conf['saved_config']), "install.sav");
		}
	}

	// RESTORE_CONFIG
	function restore_config()
	{
		if($this->my_file_exists("install.sav")) {
			$this->conf['saved_config'] = unserialize($this->file2str("install.sav"));
		}
		else $this->conf['saved_config'] = array();
	}

	// SEARCH_DIR
	function search_dir($dirs_array)
	{
		foreach ($dirs_array as $dir){  /* searches for an existent directory
																			 in an array of paths */
			if($this->my_dir_exists($dir)){
				return $dir;
			}
		}
		return "";
	}

	// SEARCH_FILE
	function search_file($dirs_array, $files_arr)
	{
		foreach ($dirs_array as $dir){  /* searches for an existent directory and file
																			 in an array of paths */
			if($this->my_dir_exists($dir)){

				foreach($files_arr as $file){
					if($this->my_file_exists($dir.$this->conf['slash'].$file)){
						return $dir;
					}
				}
			}
		}
		return "";
	}

	// SELECT_DIR
	function select_dir($msg, $default="")
	{
		while(true){
			$dir = $this->choose_dir($msg, $default);
			$dir = $this->nice_path($dir);

			if(!$this->my_dir_exists($dir)){
				if($this->my_file_exists($dir)){
					$this->msgbox("\"$dir\" must be a directory!");
				}
				else{
					$this->msgbox("Directory: \"$dir\" doesn't exist!");
				}
			}
			else{
				break;
			}
		}

		return $dir;
	}

	// SELECT_FILE
	function select_file($msg, $default="")
	{
		while(true){
			$file = $this->inputbox($msg, $default);
			$file = $this->nice_path($file);

			if(!$this->my_file_exists($file)){
				if($this->my_dir_exists($file)){
					$this->msgbox("\"$file\" must be a file!");
				}
				else{
					$this->msgbox("File: \"$file\" doesn't exist!");
				}
			}
			else{
				break;
			}
		}

		return $file;
	}

	// CHOOSE_DIR
	function choose_dir($msg, $default, $min_space_size=0, $illegal_chars=" \;\$\#\'\"")
	{
		while(true){
			$dir = $this->inputbox($msg, $default);
			$dir = $this->expand_tilde($dir);
			if(!preg_match("/^\//", $dir)){
				$this->msgbox("The path must be absolute!");
			}
			else if($illegal_chars && preg_match("/[".$illegal_chars."]/", $dir)) {

				$chars_str = implode(",", preg_split("//", trim(stripslashes($illegal_chars)), -1, PREG_SPLIT_NO_EMPTY));
				if($chars_str != "") {
					$chars_str = ": $chars_str and";
				}
				$chars_str .= " the whitespace character";

				$this->msgbox("The path contains illegal characters!\n"
					."It must not contain".addslashes($chars_str).".");
			}
			else{
				$dir = $this->nice_path($dir);
				if($min_space_size != 0) {
					$install_target = $dir;
					while(strlen($install_target) > 1 && !$this->my_dir_exists($install_target)) {
						$install_target = $this->my_dirname($install_target);
					}
					$actual_free_space = disk_free_space($install_target);
					if($actual_free_space < $min_space_size) {
						$this->install_log("Directory $dir has a free disk space: ".
								$this->size_human($actual_free_space)." , but at least ".$this->size_human($min_space_size)." is needed");
						$this->msgbox("The directory must have at least ".$this->size_human($min_space_size).
								" of free disk space.\nPlease provide another directory.");
						continue;
					}
				}
				break;
			}
		}
		return $dir;
	}

	// CHOOSE_PREFIX
	function choose_prefix($what, $default, $min_space_size=0, $illegal_chars=" \;\$\#\'\"")
	{
		return $this->choose_dir("Specify the location where to install $what", $default,
			$min_space_size, $illegal_chars);
	}

	// CHOOSE_INSTALL_PREFIX
	function choose_install_prefix($default)
	{
		if(isset($this->conf['prefix'])) {
			if(!$this->conf['interactive']) {
				return;
			}
			$default = $this->conf['prefix'];
		}
		$this->install_log("Choosing install prefix ...");

		$this->conf['prefix'] = $this->choose_prefix($this->conf['product'], $default,
				isset($this->conf['min_space_size']) ? $this->conf['min_space_size'] : 0);

		$this->conf['saved_config']['prefix'] = $this->conf['prefix'];
	}

	// USER2UID
	function user2uid($user)
	{
		if(preg_match("/^\d+$/", $user)){
			return $user;
		}
		$uid = -1;

		// try one strategy (execute id)
		if($this->my_file_exists("/usr/xpg4/bin/id")){ // first try POSIX id
			$id_out = $this->cmd2str("/usr/xpg4/bin/id -u $user 2>/dev/null");
		}
		else{
			$id_out = $this->cmd2str("id -u $user 2>/dev/null");
		}
		if(isset($id_out) && preg_match("/^\d+$/", $id_out)){
			$uid = $id_out;
		}

		// try the second strategy (look in /etc/group)
		if($uid == -1){
			if(preg_match("@^".$user.":[^\:]*:(\d+)@m", $this->file2str($this->conf['passwdfile']), $match)){
				$uid = $match[1];
			}
		}

		if($uid == -1){
			$this->on_error("Cannot find UID of user: $user");
		}

		return $uid;
	}

	// LINK_STARTUP_DAEMON
	function link_startup_daemon($program_path) /* program must understand 'start'
																								 and 'stop' arguments */
	{
		$program = basename($program_path);

		$this->install_log("Trying to link the RC initialization script to the $program ...");

		if($this->conf['uname']['sysname'] == "FreeBSD"){
			if($this->my_file_exists("/etc/rc.conf")){
				if(preg_match("/^\s*local_startup\s*=\s*\"?([^\s\#\"]+)\"?/m",
							$this->file2str("/etc/rc.conf"), $match)){
					$rc_dir = $match[1];
				}
			}
			if(empty($rc_dir)){
				$rc_dir = "/usr/local/etc/rc.d";
			}
		}
		elseif($this->conf['uname']['sysname'] == "Linux"){
			$rc_dir = "/etc/rc.d/rc5.d";
		}
		elseif($this->conf['uname']['sysname'] == "SunOS"){
			$rc_dir = "/etc/rc3.d";
		}
		elseif($this->conf['uname']['sysname'] == "Darwin"){
			$rc_dir = "/usr/local/etc/rc.d";
		}

		if(!empty($rc_dir)){
			$this->infobox("Updating system RC ...");
			if(!$this->my_dir_exists($rc_dir) || !$this->my_symlink($program_path, "$rc_dir/S$program.Zend")){
				$this->msgbox(
						"Cannot setup $program to load automatically from system RC file:\n\"$rc_dir/S$program.Zend\"\n".
						"You will have to fix this problem after the installation is complete");
			}
		}
		else{
			$this->msgbox(
					"Cannot setup $program to load automatically\n".
					"Automatic loading is not supported for this platform\n".
					"You will have to fix this problem after the installation is complete");
		}
	}

	// CONSTRUCT_CRONTAB_LINE
	function construct_crontab_line($min_frequency, $args)
	{
		if($min_frequency <= 0 || $min_frequency > 60){
			$this->on_error("Frequency must be in range between 1 and 60");
		}

		$crontab_line = "";
		for($i=0; $i<60; $i += $min_frequency){
			$crontab_line .= "$i";
			if($i < 60-$min_frequency){
				$crontab_line .= ",";
			}
		}

		$crontab_line .= " * * * *  $args";
		$this->install_log("Constructing a crontab line: $crontab_line");

		return $crontab_line;
	}


	// --------------------------
	// FILE FUNCTIONS
	// ==========================

	// MY_MOVE
	function my_move($oldname, $newname, $delete=false)
	{
		if($delete){ /* delete destination before moving */
			$this->my_delete($newname);
		}

		if($this->my_file_exists($oldname)){
			$this->install_log("Moving: $oldname to: $newname");
			return rename($oldname, $newname);
		}
		else if($this->my_dir_exists($oldname)){
			$this->my_dir_move($oldname, $newname);
		}
		else{
			$this->on_error("Cannot rename file: \"$oldname\" (it's neither a file nor a directory).");
		}
	}

	// MY_DIR_MOVE
	function my_dir_move($oldname, $newname)
	{
		if(!$this->my_dir_exists($oldname)){
			$this->on_error("File: $oldname should be a directory.");
		}
		if($this->conf['os'] == "UNIX"){
			$this->run_command("mv $oldname $newname");
		}
		else{
			$this->my_dir_copy($oldname, $newname);
			$this->my_delete($oldname);
		}
	}

	// MY_TOUCH
	function my_touch($file)
	{
		$dir = dirname($file);
		if(!$this->my_dir_exists($dir)){
			$this->my_mkdir_long($dir);
		}

		if(!$this->my_file_exists($file)){
			$this->install_log("Creating file: $file");
			if(!$this->run_silently("touch", $file)){
				$this->on_error("Cannot create file: $file");
			}
		}
	}

	// MY_COPY
	function my_copy($oldname, $newname)
	{
		if($this->my_dir_exists($newname)){
			$newname = $newname.$this->conf['slash'].basename($oldname);
		}

		$this->my_rename ($newname, "$newname.old");
		$this->my_unlink ("$newname.old"); // Unlink the old file


		if($this->my_link_exists($newname)){ /* if link is broken - delete it */
			$dest = $this->strip_symlink($newname);
			if(empty($dest)){
				$this->my_unlink($newname);
			}
		}

		if($this->my_file_exists($oldname)){
			$perms = fileperms($oldname);

			if(!$this->my_is_writable($newname)){
				$this->infobox("Cannot overwrite: $newname");
			}
			else{
				if(!$this->run_silently("copy", $oldname, $newname)){
					$this->infobox("Cannot copy: $oldname to: $newname");
				}
				else{
					$this->install_log("Copying: $oldname to: $newname");

					if(!$this->run_silently("chmod", $newname, $perms)){
						$this->infobox("Cannot change permissions of $newname to $perms");
					}
				}
			}
		}
		else if($this->my_dir_exists($oldname)){
			$this->my_dir_copy($oldname, $newname);
		}
		else{
			$this->on_error("Cannot copy file: \"$oldname\" (it's neither a file nor a directory).");
		}
	}

	// MY_DIR_COPY
	function my_dir_copy($oldname, $newname)
	{
		$this->my_mkdir($newname);
		$dir = $this->my_opendir($oldname);
		while($file = readdir($dir)){
			if($file == "." || $file == ".."){
				continue;
			}
			if($this->my_dir_exists($oldname.$this->conf['slash'].$file)
				&& $this->my_dir_exists($newname.$this->conf['slash'].$file)) {
				$this->my_dir_copy($oldname.$this->conf['slash'].$file, $newname.$this->conf['slash'].$file);
			}
			else $this->my_copy($oldname.$this->conf['slash'].$file, $newname.$this->conf['slash'].$file);
		}
		closedir($dir);
	}

	// MY_IS_WRITABLE
	function my_is_writable($file)
	{
		if(!$this->my_file_exists($file)){
			return true;
		}

		$fp = $this->run_silently("fopen", $file, "w");
		if($fp == false){
			return false;
		}
		else{
			fclose($fp);
			return true;
		}
	}

	// MY_UNLINK
	function my_unlink($file)
	{
		$this->install_log("Unlinking: $file");

		if($this->my_file_exists($file)){
			$this->run_silently("unlink", $file);
		}
	}

	// MY_RENAME
	function my_rename($from, $to)
	{
		$this->install_log("Renaming: $from to: $to");

		if($this->my_file_exists($from)){
			$this->run_silently("rename", $from, $to);
		}
	}

	// MY_SYMLINK
	function my_symlink($from, $to)
	{
		if(!$this->my_file_exists($from)){
			$this->on_error("Cannot symlink file: $from (It doesn't exist).");
		}

		$this->my_unlink($to);

		$this->install_log("Symlinking $to -> $from");
		return $this->run_silently("symlink", $from, $to);
	}

	// MY_FILE_EXISTS
	function my_file_exists($file)
	{
		if(empty($file)){
			return false;
		}
		return  ($this->run_silently("file_exists", $file) && $this->run_silently("is_file", $file)) ||
			($this->run_silently("is_link", $file) && $this->run_silently("readlink", $file));
	}

	// MY_LINK_EXISTS
	function my_link_exists($file)
	{
		return $this->my_file_exists($file) && is_link($file);
	}

	// MY_DIR_EXISTS
	function my_dir_exists($dir)
	{
		if(empty($dir)){
			return false;
		}
		return $this->run_silently("file_exists", $dir) &&
			$this->run_silently("is_dir", $dir);
	}

	// MY_DELETE
	function my_delete($file) /* recursive deletion of directories and files */
	{
		$this->install_log("Deleting: $file");

		if($this->my_file_exists($file)){
			$this->my_unlink($file);
		}
		else if($this->my_dir_exists($file)){
			if($this->conf['os'] == "UNIX"){
				$this->run_command("rm -rf $file");
			}
			else{
				$dir = $this->my_opendir($file);
				while($file_in = readdir($dir)){
					if($file_in != "." && $file_in != ".."){
						$this->my_delete($file.$this->conf['slash'].$file_in);
					}
				}
				closedir($dir);
				if($this->conf['os'] == "UNIX"){
					$this->run_silently("rmdir", $file);
				}
				elseif($this->conf['os'] == "WINDOWS"){
					$this->run_silently("rd", $file);
				}
			}
		}
	}

	// MY_MKDIR_SILENTLY
	function my_mkdir_silently($dir)
	{
		$this->install_log("Creating directory: $dir (silently)");
		return $this->my_mkdir_long($dir, true);
	}

	// MY_MKDIR_LONG
	function my_mkdir_long($dir, $silent = false)
	{
		if(!$this->my_dir_exists($dir)){
			$this->install_log("Creating (long path) directory: $dir");

			if($this->conf['os'] == "UNIX") {
				$this->run_command("mkdir -p '$dir'", "Cannot create directory: $dir");
			}
			else {
				$dir_arr = preg_split("/\/+/", $dir);
				$path = "";
				for($i=0; $i<count($dir_arr); $i++){
					$path .= $dir_arr[$i].$this->conf['slash'];
					if(!$this->my_mkdir($path, $silent)){
						return false;
					}
				}
			}
		}

		return true;
	}

	// MY_MKDIR
	function my_mkdir($dir, $silent = false)
	{
		if(!$this->my_dir_exists($dir)){
			$this->install_log("Creating directory: $dir");
			if(!$this->run_silently("mkdir", $dir)){
				if(!$silent){
					$this->on_error("Cannot create directory: $dir");
				}
				return false;
			}
		}

		return true;
	}

	// MY_DIRNAME
	function my_dirname($file)
	{
		$dirname = dirname($file);
		if($dirname == "."){
			$dirname = "";
		}

		return $dirname;
	}

	// STRIP_SYMLINK
	function strip_symlink($file)
	{
		$file_dir = $this->my_dirname($file);

		$new_file=$file;
		$count = 0;
		while($this->my_link_exists($new_file)){
			if ($count++ >= 64){
				$this->on_error("Too many levels of symbolic links for $file");
			}
			$newer_file = readlink($new_file);
			if(!$newer_file){
				$this->on_error("Unable to read symbolic link $new_file");
			}
			$new_file = $this->make_full_path($newer_file, $file_dir);
		}

		if(!is_file($new_file)){
			$this->install_log("Broken link: $file");
			return "";
		}
		return $new_file;
	}

	// IS_BROKEN_LINK
	function is_broken_link($file)
	{
		return ($this->strip_symlink($file) == "");
	}

	// MY_LIST_DIR
	function my_list_dir($path, $pattern)
	{
		$files_list = array();
		$dir = $this->my_opendir($path);
		while($file = readdir($dir)){
			if($file == "." || $file == ".."){
				continue;
			}
			if($this->my_fnmatch($pattern, $file)){
				array_push($files_list, $this->nice_path("$path/$file"));
			}
		}
		closedir($dir);

		return $files_list;
	}

	// MY_BACKUP
	function my_backup($file, $backup_link_dest=false)
	{
		$this->install_log("Trying to backup file: $file");

		if($this->my_link_exists($file)){ /* don't backup links */
			if($backup_link_dest){
				$dest = $this->strip_symlink($file);
				if(empty($dest)){
					$this->my_unlink($file);
				}
				else{
					$this->install_log("Backing-up link destination: $dest to: $dest".$this->conf['backup_suffix']);
					$this->my_copy($dest, $dest.$this->conf['backup_suffix']);
					$new_file = $dest.$this->conf['backup_suffix'];
				}
			}
			else{
				$this->install_log("Not backing-up a link: $file");
			}
		}
		elseif($this->my_file_exists($file)){
			$this->install_log("Backing-up file: $file to: ".$file.$this->conf['backup_suffix']);
			$this->my_copy($file, $file.$this->conf['backup_suffix']);
			$new_file = $file.$this->conf['backup_suffix'];
		}
		elseif($this->my_dir_exists($file)){
			$this->install_log("Backing-up directory: $file to: ".$file.$this->conf['backup_suffix']);
			$this->my_move($file, $file.$this->conf['backup_suffix'], true);
			$new_file = $file.$this->conf['backup_suffix'];
		}

		return isset($new_file) ? $new_file : null;
	}

	// MY_FNMATCH
	function my_fnmatch($pattern, $file)
	{
		for($i=0; $i<strlen($pattern); $i++) {

			if($pattern[$i] == "*") { /* star wildcard (matches all or nothing :) */
				for($c=$i; $c<max(strlen($pattern), strlen($file)); $c++) {
					if($this->my_fnmatch(substr($pattern, $i+1), substr($file, $c))) {
						return true;
					}
				}
				return false;
			}

			if($pattern[$i] == "[") { /* [] set of letters (matches something :) */

				unset($sqr_braket);
				$letter_set = array();

				for($c=$i+1; $c<strlen($pattern); $c++) {
					if($pattern[$c] != "]") {
						array_push($letter_set, $pattern[$c]);
					}
					else {
						$sqr_braket = true;
						break;
					}
				}
				if(!isset($sqr_braket)) {
					$this->on_error("No closing ']' found.");
				}
				foreach ($letter_set as $letter) {
					if($this->my_fnmatch($letter.substr($pattern, $c+1), substr($file, $i))) {
						return true;
					}
				}
				return false;
			}

			if($pattern[$i] == "?") { /* question sign wildcard (matches only one thing :) */
				continue;
			}

			if($pattern[$i] != $file[$i]) {
				return false;
			}
		}

		return true;
	}

	// MY_SCANDIR
	function my_scandir($dir)
	{
		$files = array();
		if($this->my_dir_exists ($dir)) {
			$fh = opendir($dir);
			while (false !== ($filename = readdir($fh))) {
				array_push($files, $filename);
			}
			closedir($fh);
		}
		return $files;
	}

	// FIND_FILE
	function find_file($dir, $pattern, $ret_arr=null)
	{
		$num_files = 0;
		if($this->my_dir_exists ($dir)) {
			$files = preg_grep ("/^\.+$/", $this->my_scandir ($dir), PREG_GREP_INVERT);;
			foreach ($files as $file) {
				$num_files += $this->find_file ($this->make_path($dir, $file), $pattern, $ret_arr);
			}
		}
		else if($this->my_file_exists ($dir)) {
			if($this->my_fnmatch($pattern, basename($dir))) {
				if(is_array($ret_arr)) {
					array_push ($ret_arr, $dir);
				}
				return 1;
			}
		}
		return $num_files;
	}

	// FILECMP
	function filecmp($file1, $file2)
	{
		return strcmp(file2str($file1), file2str($file2));
	}

	// FILE2STR
	function file2str($file)
	{
		if($this->my_dir_exists($file)){
			$this->on_error("Can't read \"".$file."\": It should not be a directory.");
		}

		return @file_get_contents($file);
	}

	// STR2FILE
	function str2file($str , $file)
	{
		$fp = $this->my_fopen($file, "w");
		$ret = fwrite($fp, $str);
		fclose($fp);

		return $ret;
	}

	// CMD2STR
	function cmd2str($cmd)
	{
		$fp = $this->my_popen($cmd, "r");

		$str = "";
		while(!feof($fp)){
			$str .= fread($fp, 4096);
		}
		$str = preg_replace("/^\s*|\s*$/", "", $str);

		$this->safe_pclose($fp);
		return ($str);
	}

	// FILE_WORD_WRAP
	function file_word_wrap($file)
	{
		$this->install_log("Wrapping text from file: $file");

		$tmp_file = $this->conf['tmp_dir'].$this->conf['slash'].basename($file).".wrapped";
		$this->str2file(wordwrap($this->file2str($file), $this->conf['wrap_size']), $tmp_file);
		return $tmp_file;
	}

	// MY_OPENDIR
	function my_opendir($path)
	{
		if(!$this->my_dir_exists($path)){
			$this->on_error("No such directory: $path");
		}
		$dir = $this->run_silently("opendir", $path);
		if(empty($dir)){
			$this->on_error("Cannot open directory: $path for reading");
		}

		return $dir;
	}


	// MY_FOPEN
	function my_fopen($file, $mode)
	{
		// define string for mode
		if($mode[0] == "r")      { $modestr = "reading";       }
		else if($mode[0] == "w") { $modestr = "writing";       }
		else if($mode[0] == "a") { $modestr = "appending";     }
		else                     { $modestr = "unknown mode";  }

		if(strlen($mode)>1 && $mode[1]=="b"){
			$modestr .= " (Binary)";
		}

		$this->install_log("Open file: $file for $modestr");

		$fp = $this->run_silently("fopen", $file, $mode);
		if(empty($fp)){
			$this->on_error("Cannot open file: \"$file\" for $modestr");
		}
		return $fp;
	}

	// MY_FWRITE
	function my_fwrite($file, $mode, $str)
	{
		$num_written = -1;
		$fp = $this->my_fopen($file, $mode);
		if(!empty($str)){
			$num_written = fwrite($fp, $str);
		}
		if(!fclose($fp) && $num_written<=0){
			$this->on_error("Cannot write to file: $file");
		}
		return $num_written;
	}

	// MY_POPEN
	function my_popen($cmd, $mode, $need_loging=true)
	{
		if($need_loging){
			$this->install_log("Executing: $cmd");
		}

		$fp = popen($cmd, $mode);
		if($fp == false){
			$this->on_error("Cannot execute command: $cmd");
		}
		return $fp;
	}

	// PUSHD
	function pushd($dir)
	{
		$this->install_log("Saving current working directory");

		$cwd = getcwd();
		if(empty($cwd)){
			$this->on_error("Unable to store the current working directory.");
		}
		if(!isset($this->conf['cwd'])) {
			$this->conf['cwd'] = array();
		}
		array_push($this->conf['cwd'], $cwd);

		$this->my_chdir($dir);
	}

	// POPD
	function popd()
	{
		$this->install_log("Restoring current working directory");

		if(!isset($this->conf['cwd'])) {
			$error = "Current directory is not stored";
		}
		$cwd = array_pop($this->conf['cwd']);
		if(!chdir($cwd)){
			$error = "Can't change directory to: $cwd";
		}
		if(isset($error)) {
			$this->on_error("Unable to restore the current working directory.\n$error");
		}
	}

	// MY_CHDIR
	function my_chdir($dir)
	{
		$this->install_log("Changing directory to: $dir");
		if(!chdir($dir)){
			$this->on_error("Unable to change directory to: $dir");
		}
	}

	// GET_FILE_TREE
	function get_file_tree($dir, &$file_tree)
	{
		$totalsize = 0;
		$file_tree = array();
		$this->get_file_tree_a($dir, $file_tree, $totalsize);
		return $totalsize;
	}

	function get_file_tree_a($path, &$file_tree, &$totalsize) /* returns number of files */
	{
		$dir = $this->my_opendir($path);
		$num_files = 0;

		while($file = readdir($dir)){
			if($file != "." && $file != ".."){
				if($this->my_file_exists("$path/$file")){
					$size = filesize($path.$this->conf['slash'].$file);
					$file_tree[$path.$this->conf['slash'].$file] = $size;
					$totalsize += $size;
				}
				elseif($this->my_dir_exists($path.$this->conf['slash'].$file)){
					if($this->get_file_tree_a($path.$this->conf['slash'].$file, $file_tree, $totalsize)==0){
						$file_tree[$path.$this->conf['slash'].$file] = 0; /* if directory is empty put it into
																																 components array */
					}
				}
				$num_files ++;
			}
		}

		closedir($dir);
		return $num_files;
	}


	// ------------------------------
	// DIALOG FUNCTIONS
	// ==============================

	// MY_READLINE
	function my_readline($msg = "[To continue, press Enter]")
	{
		print($msg);
		flush();
		if($this->conf['interactive']) {
			exec("read OUTPUT\necho \$OUTPUT", $output);
		}
		print ("\n");
		return isset($output[0]) ? $output[0] : "";
	}

	// RUNDIALOG
	function rundialog($mode, $msg)
	{
		$command = $this->conf['dialog'];
		switch($mode)
		{
			case 'textbox':
				$command .= " --$mode \"$msg\" 0 0";
				break;

			case 'msgbox':
			case 'inputbox':
			case 'passwordbox':
			case 'yesno':
			case 'infobox':
			case 'menu':
				if($mode == 'infobox') {
					$command .= " --sleep 1 ";
				}
				// remove ascii escape sequences(not suitable for dialog)
				$msg = preg_replace('/\\033[0-9;\[]*m/', '', $msg);
				$command .= " --$mode \"$msg\n\" 0 0";
				break;

			default:
				print("Unsupported dialog option\n");
				return('');
		}

		$args = func_get_args();
		array_shift($args);
		array_shift($args);
		if (sizeof($args) != 0) {
			$command .= ' "' . join(', "', $args) . '"';
		}

		$rc = $this->dialog_execute($command);
		if((($mode == 'passwordbox') || ($mode == 'inputbox')) && ($rc>0)) {
			// Cancel pressed.
			return -1;
		}

		$dialog_output = $this->dialog_get_output();
		if(preg_match ('/Error opening terminal/', $dialog_output)) {
			unset($this->conf['dialog']);
			$this->on_error($this->conf['dialog_error']);
		}

		return($mode == 'yesno' ? 1 - $rc : $dialog_output);
	}

	// MSGBOX
	function msgbox($msg)
	{
		$this->install_log("MSGBOX: \n\"$msg\"");

		if(!empty($this->conf['dialog'])){
			if(defined('DBUILDER')){
				$this->conf['dialog']->msgbox($this->conf['title'], $msg);
			}
			else{
				$this->rundialog("msgbox", "\n$msg");
			}
		}
		else{
			print("\n$msg\n");
			$this->my_readline();
		}
	}

	// INPUTBOX
	function inputbox($msg, $init="", $accept_empty=false)
	{
		$this->install_log("ASK: $msg");
		do
		{
			if(!empty($this->conf['dialog'])){
				while(true) {
					if(defined('DBUILDER')){
						$input = $this->conf['dialog']->inputbox($this->conf['title'], $msg, $init, OK|CANCEL);
					}
					else{
						$input = $this->rundialog("inputbox", "\n$msg", $init);
					}
					if($input == -1){
						$this->on_abort();
						continue;
					}
					break;
				}
			}
			else{
				// replace last ':' if it exits and trim :
				$msg = preg_replace("/\:\s*$/", "", $msg);

				if($accept_empty){
					$init = "Press ENTER to leave empty";
				}

				$input = $this->my_readline( empty($init) ? "$msg: " : "$msg [$init]: " );
				if($input == "" && !$accept_empty){
					$input = $init;
				}
			}

			$this->install_log("INPUTBOX: $input");
		}
		while((empty($input) || preg_match("/^\s*$/", $input)) && !$accept_empty);

		$input = trim($input);        /* trim the line */
		$input = addslashes($input);  /* add the back slashes to illegal characters */

		return $input;
	}

	// INPUTBOX_BOUNDED
	function inputbox_bounded($msg, $what, $min_chars, $max_chars, $init="", $accept_empty=false)
	{
		while(true)
		{
			$output = $this->inputbox($msg, $init, $accept_empty);

			if(strlen($output) < $min_chars) {
				$this->msgbox("A $what should contain at least $min_chars characters. Please try again.");
				continue;
			}
			if(strlen($output) > $max_chars) {
				$this->msgbox("A $what should contain at most $max_chars characters. Please try again.");
				continue;
			}

			break;
		}

		return $output;
	}

	// YESNOBOX
	function yesnobox($msg, $default="YES", $abort_on_no=false)
	{
		$this->install_log("YESNO: $msg");

		$answer = false;

		while(true) {
			if(!empty($this->conf['dialog'])){
				if(defined('DBUILDER')){
					$answer = $this->conf['dialog']->yesnobox($this->conf['title'], $msg, YES|NO);
				}
				else{
					$answer = $this->rundialog("yesno", "\n$msg");
				}
			}
			else{
				print("\n$msg\n");
				do{
					$output = $this->my_readline("Answer (yes or no) [$default]: ");
					if(preg_match("/^\s*$/", $output)){
						$output = $default;
					}
					if(strcasecmp($output, "yes") == 0){
						$answer = true;
						break;
					}
					if(strcasecmp($output, "no") == 0){
						$answer = false;
						break;
					}
				}while(true);
			}

			if($abort_on_no && !$answer) {
				$this->on_abort();
				continue;
			}

			print("\n");
			break;
		}

		$result = $answer ? "YES" : "NO";
		$this->install_log("User choose: $result");
		return $answer;
	}

	// INFOBOX
	function infobox($msg)
	{
		$this->install_log("INFOBOX: \n\"$msg\"");

		if(!empty($this->conf['dialog'])){
			if(defined('DBUILDER')){
				$this->conf['dialog']->infobox($this->conf['title'], $msg);
			}
			else {
				$this->rundialog("infobox", "\n$msg");
			}
		}
		else{
			print("\n$msg\n");
		}
	}

	// PASSWDBOX
	function passwdbox($msg)
	{
		do{
			if(!empty($this->conf['dialog'])){
				while(true) {
					if(defined('DBUILDER')){
						$passwd = $this->conf['dialog']->passwordbox($this->conf['title'], $msg, "", OK|CANCEL);
					}
					else{
						$passwd = $this->rundialog("passwordbox", $msg);
					}
					if($passwd == -1){
						$this->on_abort();
						continue;
					}
					break;
				}
			}else{
				print("\n$msg");
				exec("stty -echo");
				$passwd = $this->my_readline("");
				exec("stty echo");
			}
		}
		while(empty($passwd) || preg_match("/^\s*$/", $passwd));

		$this->install_log("PASSWDBOX: ". preg_replace("/\w/", "X", $passwd));

		return($passwd);
	}

	// PASSWORD_CHECK
	function password_check($passwd, $num_chars=4)
	{
		if(strlen($passwd) < $num_chars){
			$this->msgbox("A password should contain at least $num_chars characters. Please try again.");
			return false;
		}
		if((urlencode($passwd) != $passwd) or (strpos($passwd, "_") !== false)){
			$this->msgbox("Illegal password. The password may contain only:\n\n"
			."a) The alphanumeric characters 'a' through 'z',\n   'A' through 'Z' and '0' through '9'\n"
			."b) The special characters '.' and '-'\n\n"
			."Please try again.\n");
			return false;
		}
		return true;
	}

	// PASSWORD_GET
	function password_get($msg="", $num_chars=4)
	{
		while(true) {
			$passwd = $this->passwdbox("$msg\n\nPassword: ");
			if(!$this->password_check($passwd, $num_chars)) {
				continue;
			}
			break;
		}
		return $passwd;
	}

	// PASSWORD_CHOOSE
	function password_choose($msg="", $num_chars=4)
	{
		while(true){
			$passwd = $this->password_get($msg, $num_chars);
			$checkpasswd = $this->passwdbox("Verify the password: ");

			if(empty($this->conf['dialog'])){
				print("\n");
			}

			if(strcmp($passwd, $checkpasswd) == 0){
				break;
			}
			$this->msgbox("Passwords did not match, please try again.");
		}
		return $passwd;
	}

	// MENUBOX
	function menubox($msg, $menu_arr_orig, $allow_blank=false, $abort_msg="", $default=null)
	{
		$this->install_log("MENU: $msg\n:". var_export($menu_arr_orig, true));

		$temp_hash = array();
		$menu_arr = array();
		$index = 1;
		foreach($menu_arr_orig as $key=>$val){
			if($default != null && $default == $key) {
				$default = $index;
			}
			$temp_hash[$index] = $key;
			$menu_arr[$index] = $val;
			$index ++;
		}

		if(!empty($this->conf['dialog'])){

			if(defined('DBUILDER')){

				while(true) {
					$choice = $this->conf['dialog']->menubox($this->conf['title'], $msg, $menu_arr, "", OK|CANCEL);
					if($choice == -1){
						$this->on_abort();
						continue;
					}
					break;
				}
			}
			else{

				$menu_params = "";
				$item_len = 0;
				foreach ($menu_arr as $tag => $menu_item){
					$menu_item = preg_replace("/\"/", "'", $menu_item);
					$menu_params .= " \"$tag\"\t\"$menu_item\"";

					if(strlen($menu_item) > $item_len){
						$item_len = strlen($menu_item);
					}
				}
				$msg = preg_replace("/\"/", "'", $msg);

				if($item_len > strlen($msg)-6){ /* message must be longer than items */
					$len_add = $item_len - strlen($msg);
					$len_add = (strlen($msg)+10 > $this->conf['wrap_size'])?
						$this->conf['wrap_size'] - strlen($msg) : $len_add + 10;
					$msg .= str_repeat(" ", $len_add);
				}

				$msg = wordwrap($msg, $this->conf['wrap_size']);

				while(true) {
					$choice = $this->rundialog("menu", "\n$msg", "\"".count($menu_arr)." $menu_params\"");

					if(!$choice) {
						$this->on_abort($abort_msg);
						continue;
					}
					$choice = $temp_hash[$choice];
					break;
				}
			}
		}
		else{
			$msg = $this->add_dashes($msg);

			$count_lines = 0;
			$str = "\n$msg\n\n";
			foreach ($menu_arr as $tag => $menu_item){
				$str .= "[$tag] $menu_item\n";
				$count_lines ++;
			}

			do
			{
				if($count_lines > 20){
					$tmp_file = $this->make_path($this->conf['tmp_dir'], "tmp.scroll");
					$this->str2file($str, $tmp_file);
					$this->show_file($tmp_file, false);
				}
				else{
					print($str);
				}

				while(true){

					if($allow_blank){
						$quest = "\nPlease select [Press ENTER to leave empty]: ";
						if($count_lines > 20){
							$quest = "\nPlease select [S - show again, ENTER - leave empty]: ";
						}
					} else {
						$quest = "\nPlease select ";
						$quest .= (($default == null) ? ": " : "[$default]: ");
						if($count_lines > 20){
							$quest = "\nPlease select [S - show again]: ";
						}
					}

					$output = $this->my_readline($quest);
					if($output == "S"){
						break;
					}

					if(empty($output) && $default != null) {
						$output = $default;
					}

					if(empty($output) && $allow_blank){ /* if ENTER was pressed */
						$choice = "";
						break;
					}
					if(isset($menu_arr[$output])){ /* if this value exists */
						$choice = $temp_hash[$output];
						break;
					}
					else {
						if(!empty($output)) {
							$this->infobox("Illegal input: $output");
						}
					}
				}
			}
			while(!isset($choice));

			print("\n");
		}

		$this->install_log("User's choice: $choice");
		return $choice;
	}

	// SHOW_FILE
	function show_file($file, $clean=true)
	{
		if(!$this->conf['interactive']) {
			return;
		}
		$file = $this->file_word_wrap($file);

		if(!empty($this->conf['dialog'])){
			if(defined('DBUILDER')){
				$this->conf['dialog']->textbox($file, $this->conf['title'], OK);
			}
			else {
				$this->rundialog("textbox", $file);
			}
		}
		else{
			$this->clrscr();
			if($this->conf['os'] == "UNIX"){
				popen("cat $file | more", "w");
			}
			else{
				popen("more $file");
			}
			if($clean){
				$this->clrscr();
			}
		}
	}

	// PROGRESS_BAR_START
	function progress_bar_start($seconds, $message, $message_tty=null)	/* start different process
																																				 with dialog progress bar */
	{
		$message = "\n\n".$message;
		if(!$message_tty){
			$message_tty = $message;
		}
		putenv("TIME=$seconds"); /* set maximal progress time */
		putenv("MESSAGE=$message");
		putenv("MESSAGE_TTY=$message_tty");
		putenv("TITLE=".$this->conf['title']); /* for DBUILDER dialogs */

		if(!empty($this->conf['dialog'])){
			if(defined('DBUILDER')){
				putenv("DIALOG=DBUILDER");
			}
			else putenv("DIALOG=".$this->conf['dialog']); /* set path to dialog */
		}

		$this->conf['progress_pid_file'] = $this->conf['tmp_dir'].$this->conf['slash']."progress_bar.pid";
		$this->conf['progress_fatal_file'] = $this->conf['tmp_dir'].$this->conf['slash']."kill_progress_bar";

		putenv("PID_FILE=".$this->conf['progress_pid_file']);
		putenv("FATAL_FILE=".$this->conf['progress_fatal_file']);

		$dspec = array(
				0 => array("pipe", "r"),
				2 => array("pipe", "w")
				);

		$this->conf['progress_fp'] = proc_open("./php -q ./progress_bar.php", $dspec, $pipes);
		$this->conf['progress_pipes'] = $pipes;
		$this->conf['progress_bar_started'] = true;
	}

	// PROGRESS_BAR_STOP
	function progress_bar_stop() /* stop the progress bar by killing it */
	{
		while(!file_exists($this->conf['progress_pid_file'])); // Wait for process to create a file
		sleep(1);

		// Old style kill:
		//$pid = $this->file2str($this->conf['progress_pid_file']);
		//posix_kill($pid, 9);

		// New style kill:
		touch($this->conf['progress_fatal_file']);

		proc_close($this->conf['progress_fp']);

		if(empty($this->conf['dialog'])){
			print("\n");
		}
		$this->conf['progress_bar_started'] = false;

		sleep(1);
	}

	// DIALOG_CALC_MAX_SIZE
	function dialog_calc_max_size()
	{
		if(empty($this->conf['dialog']) || defined('DBUILDER')){
			return -1;
		}

		$this->dialog_execute($this->conf['dialog']." --print-maxsize");

		$out = $this->dialog_get_output();

		if(preg_match("/.*:\s(.*),\s(.*)$/m", $out, $match)){
			$this->conf['max_y'] = $match[1];
			$this->conf['max_x'] = $match[2];
		}
	}

	// DIALOG_EXECUTE
	function dialog_execute($dialog_cmd)
	{
		if(empty($this->conf['dialog'])){
			$this->install_log("Tried to run a dialog utility, while it doesn't exist!");
			return -1;
		}

		$fd = $this->my_popen("$dialog_cmd 2>". $this->conf['dialog_tmp'], "w", false);
		if(empty($fd)){
			$this->on_error($this->conf['dialog_error']);
		}
		$status = $this->safe_pclose($fd);
		return $status;
	}

	// DIALOG_GET_OUTPUT
	function dialog_get_output()
	{
		if(!empty($this->conf['dialog'])){
			return $this->file2str($this->conf['dialog_tmp']);
		}
		return false;
	}


	// -----------------------------
	// WEBSERVER FUNCTIONS
	// =============================

	// WEBSERVER_SUPPORTED_CHECK
	function webserver_supported_check($name)
	{
		if(empty($this->conf['supported_webservers'])){
			return true;
		}
		else{
			return array_search($name, $this->conf['supported_webservers']);
		}
	}

	// WEBSERVER_CHOOSE
	function webserver_choose($errmsg="")
	{
		if($this->conf['webserver'] != "webserver") {
			return true;
		}

		if(empty($this->conf['supported_webservers'])){
			return false;
		}

		if(count($this->conf['supported_webservers']) == 1){
			if($this->yesnobox("Are you using ".$this->conf['supported_webservers'][1]." webserver?")){
				$this->conf['webserver'] = $this->conf['supported_webservers'][1];
			}
			else{
				if(empty($errmsg)){
					$errmsg = "For now ".$this->conf['product']." supports only ".
						$this->conf['supported_webservers'][1]. " on ". $this->conf['uname']['sysname'];
				}
				$this->on_abort($errmsg, true);
			}
		}
		else{
			$rc = $this->menubox("Choose one of supported webservers:", $this->conf['supported_webservers'], false, $errmsg);
			$this->conf['webserver'] = $this->conf['supported_webservers'][$rc];
		}
	}

	// WEBSERVER_ROOT_GUESS
	function webserver_root_guess()
	{
		if(empty($this->conf['webserver_root_dir'])){

			$webserver_root = "";
			if($this->conf['webserver'] == "Apache"){
				$webserver_root = $this->apache_root_guess();
			}
			else if($this->conf['webserver'] == "Zeus" && $this->my_dir_exists("/usr/local/zeus")) {
				$webserver_root = "/usr/local/zeus";
			}
			else if($this->conf['webserver'] == "SunONE" && $this->my_dir_exists("/opt/SUNWwbsvr")) {
				$webserver_root = "/opt/SUNWwbsvr";
			}

			$word = empty($webserver_root) ? "enter" : "confirm";
			$this->conf['webserver_root_dir'] = $this->select_dir("Please $word your ".$this->conf['webserver'].
					" root directory", $webserver_root);

			if($this->conf['webserver'] == "Zeus"){
				$this->conf['zeus_root'] = $this->conf['webserver_root_dir'];
			}
			else if($this->conf['webserver'] == "SunONE") {
				$this->conf['sunone_root'] = $this->conf['webserver_root_dir'];
			}
		}

		return $this->conf['webserver_root_dir'];
	}

	// WEBSERVER_DOCROOT_GUESS
	function webserver_docroot_guess()
	{
		if(isset($this->conf['webserver_doc_root']) && $this->my_dir_exists($this->conf['webserver_doc_root'])){
			return $this->conf['webserver_doc_root'];
		}

		$this->install_log("Trying to guess ".$this->conf['webserver']." docroot ...");

		$default = "";
		if($this->conf['webserver'] == "Apache"){
			$default = $this->apache_docroot_guess();
		}
		elseif($this->conf['webserver'] == "Zeus"){
			$default = $this->zeus_docroot_guess();
		}
		elseif($this->conf['webserver'] == "SunONE"){
			$default = $this->sunone_docroot_guess();
		}
		else{
			$default = "/var/www/html";
		}

		$manually_set = false;
		if(!isset($this->conf['webserver_doc_root'])) {
			$manually_set = true;
		}
		else if(!$this->my_dir_exists($this->conf['webserver_doc_root'])) {
			$this->msgbox("Directory:\n".$this->conf['webserver_doc_root']."\ndoesn't exist!");
			$manually_set = true;
		}

		if($manually_set) {
			$word = empty($default) ? "enter" : "confirm";
			$this->conf['webserver_doc_root'] = $this->select_dir(
					"Please $word the location of your ".$this->conf['webserver']." document root directory.\n".
					"This is the location where your web documents are stored", $default);

		}

		if($this->conf['webserver'] == "Apache"){
			$this->conf['apache_doc_root'] = $this->conf['webserver_doc_root'];
		}
		elseif($this->conf['webserver'] == "Zeus"){
			$this->conf['zeus_doc_root'] = $this->conf['webserver_doc_root'];
		}

		$this->install_log("Guessed: ".$this->conf['webserver_doc_root']);

		return $this->conf['webserver_doc_root'];
	}

	// WEBSERVER_GET_USER
	function webserver_get_user($msg="")
	{
		$this->install_log("Trying to guess ".$this->conf['webserver']."'s UID");

		$user = "";
		if($this->conf['webserver'] == "Apache"){
			$user = $this->apache_user_guess();
		}
		else if($this->conf['webserver'] == "Zeus"){
			$user = $this->zeus_user_guess();
		}
		else if($this->conf['webserver'] == "SunONE"){
			$user = $this->sunone_user_guess();
		}

		$word = empty($user) ? "enter" : "confirm";
		if(empty($msg)){
			$msg = "Please $word the username that your PHP scripts run as under ".$this->conf['webserver'];
		}

		$default = $user;
		while(true){
			$user = $this->inputbox($msg, $default);

			if(!preg_match("#^\s*$user\:#m", $this->file2str($this->conf['passwdfile']))){
				if(!$this->yesnobox("Cannot find user: '$user'\nDo you want to try again?")){
					$this->on_error("The installation cannot continue.");
				}
			}
			else{
				break;
			}
		}

		if($this->conf['webserver'] == "Apache"){
			$this->conf['apache_user'] = $user;
		}

		return $user;
	}

	// WEBSERVER_RESTART
	function webserver_restart()
	{
		$this->install_log("Trying to restart ".$this->conf['webserver']." ...");

		$restarted=false;

		if($this->conf['webserver'] == "Apache"){
			$restarted = $this->apache_restart();
		}
		elseif($this->conf['webserver'] == "Zeus"){
			$restarted = $this->zeus_restart();
		}
		elseif($this->conf['webserver'] == "SunONE"){
			$restarted = $this->sunone_restart();
		}
		else{
			$this->warn("Don't know how to restart ". $this->conf['webserver']);
		}

		if($restarted){
			$this->msgbox($this->conf['webserver']." has restarted successfully.");
			return true;
		}

		$this->msgbox("Installation failed to restart your Web server,\nplease restart it manually.");
		return false;
	}


	// -----------------------------
	// APACHE FUNCTIONS
	// =============================

	// APACHE_ROOT_SET
	function apache_root_set($apache_root)
	{
		$this->conf['apache_root_dir'] = $apache_root;
		$this->conf['webserver_root_dir'] = $apache_root;
	}

	// APACHE_CONF_SET
	function apache_conf_set($apache_conf)
	{
		$this->conf['apache_conf_dir'] = $apache_conf;
		$this->conf['apache_conf_file'] = $this->make_path($apache_conf, "httpd.conf");
	}

	// APACHE_EXEC_SET
	function apache_exec_set($apache_exec)
	{
		$this->conf['apache_exec_dir'] = $apache_exec;
		$this->conf['apache_exec_file'] = $this->make_path($apache_exec, "httpd");
	}

	// APACHE_LIBEXEC_SET
	function apache_libexec_set($apache_libexec)
	{
		$this->conf['apache_libexec_dir'] = $apache_libexec;
	}

	// APACHE_DOCROOT_SET
	function apache_docroot_set($apache_docroot)
	{
		$this->conf['apache_doc_root'] = $apache_docroot;
		$this->conf['webserver_doc_root'] = $apache_docroot;
	}

	// APACHE_CONF_READ
	function apache_conf_read()
	{
		$this->apache_conf_guess();

		if(!empty($this->conf['apache_conf_cont'])) {
			return $this->conf['apache_conf_cont'];
		}

		if(empty($this->conf['apache_conf_cont']) && $this->my_file_exists ($this->conf['apache_conf_file'])) {

			$this->conf['apache_conf_cont'] = "";
			$this->install_log("Opening Apache configuration file.");

			$apache_cont = $this->file2str($this->conf['apache_conf_file']);
			foreach(explode("\n", $apache_cont) as $line) {

				if(preg_match("@^\s*Include\s+\"?([^\s\"]+)\"?@", $line, $match)) { /* insert include file */

					$include_path = $match[1];

					/* Guess Apache root directory */
					if(empty($this->conf['apache_root_dir']) && preg_match("#^\s*ServerRoot\s+\"?([^\s\"]+)\"?#m", $apache_cont, $match)){
						$this->conf['apache_root_dir'] = $match[1];
					}
					if(empty($this->conf['apache_root_dir'])) {
						$this->apache_guess_from_binary();
					}

					$include_path = $this->make_full_path($include_path,
							isset($this->conf['apache_root_dir']) ? $this->conf['apache_root_dir'] : "");

					$include_files_arr = array();
					if(preg_match("@\*@", $include_path)) {
						$include_files_arr = $this->my_list_dir($this->my_dirname($include_path), basename($include_path));
					}
					else if($this->my_dir_exists($include_path)) {
						$include_files_arr = $this->my_list_dir($include_path, "*");
					}
					else {
						$include_files_arr[0] = $include_path;
					}

					foreach ($include_files_arr as $include_file) {
						$this->install_log("Found \"Include\" directive in ".$this->conf['apache_conf_file'].": $include_file");
						if($this->my_file_exists($include_file)) {
							$this->conf['apache_conf_cont'] .= "# zend_open<$include_file>\n";
							$this->conf['apache_conf_cont'] .= $this->file2str($include_file);
							$this->conf['apache_conf_cont'] .= "# zend_close<$include_file>\n";
						}
						else {
							$this->install_log("File: $include_file doesn't exist");
						}
					}
				}

				$this->conf['apache_conf_cont'] .= "$line\n";
			}
		} else {
			$this->conf['apache_conf_cont']="";
		}

		return $this->conf['apache_conf_cont'];
	}

	// APACHE_CONF_CLOSE
	function apache_conf_close()
	{
		if(empty($this->conf['apache_conf_cont'])){
			return false;
		}
		$this->install_log("Closing Apache configuration file.");

		if(preg_match_all("@# zend_open<(.*)>\n(.*)# zend_close<\\1>\n@Us", $this->conf['apache_conf_cont'], $match)) {
			for($i=0; $i<count($match[0]); $i++) {
				if(strcmp($match[2][$i], $this->file2str($match[1][$i])) != 0) { /* If changed this file - backup the original */
					$this->my_backup($match[1][$i], true);
				}
				$this->str2file($match[2][$i], $match[1][$i]); /* Write all includes to their destinations */
			}
		}

		$this->conf['apache_conf_cont'] = preg_replace("@# zend_open<(.*)>\n.*# zend_close<\\1>\n@Us", "",
				$this->conf['apache_conf_cont']);
		$this->str2file($this->conf['apache_conf_cont'], $this->conf['apache_conf_file']);

	}

	// IS_USING_APACHE
	function is_using_apache($silent = false)
	{
		if($this->conf['webserver'] == "Apache"){
			return true;
		}
		if($this->conf['webserver'] != "webserver"){
			return false;
		}

		if(isset($this->conf['using_apache'])){
			return ($this->conf['using_apache'] == "YES") ? true: false;
		}

		if($silent) {
			return false;
		}

		if($this->yesnobox("Are you using Apache Web server?")){
			$this->conf['using_apache'] = "YES";
			$this->conf['webserver'] = "Apache";
			return true;
		}

		$this->conf['using_apache'] = "NO";
		return false;
	}

	// APACHE_CONF_GUESS
	function apache_conf_guess($silent = false) /* guess configuration file */
	{
		if(!$this->is_using_apache($silent)){
			return false;
		}

		if(!empty($this->conf['apache_conf_file'])){
			return $this->conf['apache_conf_file'];
		}

		$this->install_log("Trying to guess Apache configuration file ...");

		$possible_configs = array(
				"httpd.conf",   /* Regular */
				"httpsd.conf",  /* For Apache compiled with SSL support */
				"httpd2.conf",   /* Some systems (like Mandrake 9.1) called so to Apache 2 configuration file */
				"apache.conf",   /* Some systems (like Mandrake 9.1)*/
				"apache2.conf" /* Some systems (like Mandrake 9.1)*/
				);

		if(!empty($this->conf['apache_conf_dir'])){
			$conf_default = $this->conf['apache_conf_dir'];
		}
		else if(isset($this->conf['apache_root_dir'])) {
			$conf_default = $this->make_path($this->conf['apache_root_dir'], "conf");

			$found = false;
			foreach ($possible_configs as $pos_config) {
				if($this->my_file_exists("$conf_default/$pos_config")) {
					$found = true;
					break;
				}
			}
			if(!$found) {
				unset($conf_default);
			}
		}

		$apache_conf_default = array( /* default paths to config dir */
				1 =>  "/usr/local/apache/conf",
				2 =>  "/usr/local/apache2/conf",
				3 =>  "/etc/httpd/conf",
				4 =>  "/usr/local/apache/etc/",
				5 =>  "/usr/local/apache2/etc/",
				6	=>	"/usr/local/etc/apache/"
				);

		if(empty($conf_default)){
			$conf_default = $this->search_file($apache_conf_default, $possible_configs);
		}

		if(isset($this->conf['default_apache_conf_dir'])) {
			$conf_default = $this->conf['default_apache_conf_dir'];
		}

		if(!$silent) {
			while($path = $this->choose_dir("Specify the Apache configuration directory", $conf_default))
			{
				foreach ($possible_configs as $pos_config) {
					if($this->my_file_exists("$path/$pos_config")) {
						$this->conf['apache_conf_file'] = $this->nice_path("$path/$pos_config");
						break;
					}
				}
				if(!(isset($this->conf['apache_conf_file'])) || !$this->my_file_exists($this->conf['apache_conf_file'])) {
					$this->msgbox("Cannot find Apache configuration in the specified directory. Please try again.");
				}
				else break;
			}
		}
		else {
			$this->conf['apache_conf_file'] = $this->make_path($conf_default, "httpd.conf");
		}

		if(@file_exists($path))
		{
			$this->conf['apache_conf_dir'] = $path;
			$this->conf['apache_root_dir'] = dirname($path);
			$this->conf['saved_config']['default_apache_conf_dir'] = $path;
			$this->install_log("Found: $this->conf['apache_conf_file']");
		}

		if(@isset($this->conf['apache_conf_file'])) {
			return $this->conf['apache_conf_file'];
		}
	}

	// APACHE_EXEC_GUESS
	function apache_exec_guess($silent = false) /* guess exec. dir */
	{
		if(!$this->is_using_apache($silent)){
			return false;
		}

		if(!empty($this->conf['apache_exec_file'])){
			return $this->conf['apache_exec_file'];
		}

		$this->install_log("Trying to guess Apache executable file ...");

		$possible_httpds = array(
				"httpd",	/* Regular */
				"httpsd", /* Compiled with SSL support */
				"httpd2", /* Some systems (like Mandrake 9.1) called to Apache 2 by this name */
				"apache",  /* Debian style */
				"apache2"
				);

		$possible_apachectls = array(
				"apachectl",  /* Regular */
				"httpdctl",
				"httpsdctl",   /* For Apache compiled with SSL support */
				"apache2ctl"
				);

		if(!empty($this->conf['apache_exec_dir'])){
			$exec_default = $this->conf['apache_exec_dir'];
		}
		else if(isset($this->conf['apache_root_dir'])) {
			$exec_default = $this->make_path($this->conf['apache_root_dir'], "bin");

			$found = false;
			foreach ($possible_httpds as $pos_httpd) {
				if($this->my_file_exists("$exec_default/$pos_httpd")) {
					$found = true;
					break;
				}
			}
			if(!$found) {
				unset($exec_default);
			}
		}

		$apache_exec_default = array( /* default paths to bin dir */
				1 =>  "/usr/local/apache/bin",
				2 =>  "/usr/local/apache2/bin",
				3 =>  "/usr/sbin",
				4 =>  "/usr/bin",
				5 =>  "/usr/local/sbin",
				6 =>  "/usr/local/bin"
				);

		if(empty($exec_default)){
			$exec_default = $this->search_file($apache_exec_default, $possible_httpds);
		}

		while($path = $this->choose_dir("Specify the Apache bin directory", $exec_default))
		{
			foreach ($possible_httpds as $pos_httpd) {
				if($this->my_file_exists("$path/$pos_httpd")) {
					$this->conf['apache_exec_file'] = $this->nice_path("$path/$pos_httpd");
					break;
				}
			}
			foreach ($possible_apachectls as $pos_apachectl) {
				if($this->my_file_exists("$path/$pos_apachectl")) {
					$this->conf['apachectl_path'] = $this->nice_path("$path/$pos_apachectl");
					break;
				}
			}
			if(!(isset($this->conf['apache_exec_file'])) || !$this->my_file_exists($this->conf['apache_exec_file'])) {
				$this->msgbox("Cannot find Apache executable in the specified directory. Please try again.");
			}
			else break;
		}


		$this->conf['apache_exec_dir'] = $path;
		if(!isset($this->conf['apachectl_path']) || !$this->my_file_exists($this->conf['apachectl_path'])){
			$this->warn("Cannot find apachectl.");
		}
		$this->install_log("Found httpd: ".$this->conf['apache_exec_file']);
		$this->install_log("Found apachectl: ".$this->conf['apachectl_path']);
		return $this->conf['apache_exec_file'];
	}

	// APACHE_GET_VIRTUAL_HOST
	function apache_get_virtual_host($default_host=null)
	{
		$this->apache_guess_from_conf();
		$host = $this->conf['uname']['nodename'];

		if($default_host) {
			$host = $default_host;
		}

		if(@is_array($this->conf['apache_vhosts'])) {
			foreach ($this->conf['apache_vhosts'] as $key => $arr) {
				if($arr['docroot'] == $this->conf['apache_doc_root'] && $arr['port'] != "443") {
					$found_key = $key;
					break;
				}
			}
		}

		if(isset($found_key)) {
			if(isset($this->conf['apache_vhosts'][$found_key]["server"])) {
				$host = $this->conf['apache_vhosts'][$found_key]["server"];
			}
			else if(isset($this->conf['apache_vhosts'][$found_key]["ip"]) && preg_match("/^[\d\.]+$/", $this->conf['apache_vhosts'][$found_key]["ip"])) {
				$host = $this->conf['apache_vhosts'][$found_key]["ip"];
			}
			if(!preg_match("/\:\d+$/", $host)) {
				$host .= ":".$this->conf['apache_vhosts'][$found_key]["port"];
			}
		}
		return $host;
	}

	// APACHE_LIBEXEC_GUESS
	function apache_libexec_guess()
	{
		if(!$this->is_using_apache()){
			return false;
		}

		if(!empty($this->conf['apache_libexec_dir'])){
			return $this->conf['apache_libexec_dir'];
		}

		$this->apache_guess_from_conf();

		$this->conf['apache_libexec_dir'] = $this->select_dir(
				"Please specify the location of your Apache modules\n(usually: HTTPD_ROOT/libexec):",
						$this->make_path($this->conf['apache_root_dir'], "libexec"));

		$this->install_log("Found Apache libexec directory at: ".$this->conf['apache_libexec_dir']);
	}

	// APACHE_ROOT_GUESS
	function apache_root_guess()
	{
		if(!$this->is_using_apache()){
			return false;
		}

		if(!empty($this->conf['apache_root_dir'])){
			return $this->conf['apache_root_dir'];
		}

		$this->apache_guess_from_conf();
		if(empty($this->conf['apache_root_dir'])){
			$this->apache_guess_from_binary();
		}

		if(empty($this->conf['apache_root_dir'])){ /* set default apache root directory */
			$this->conf['apache_root_dir'] = $this->nice_path($this->conf['apache_conf_dir']."../");
		}

		return $this->conf['apache_root_dir'];
	}

	// APACHE_DOCROOT_GUESS
	function apache_docroot_guess()
	{
		if(!$this->is_using_apache()){
			return false;
		}

		if(!empty($this->conf['apache_doc_root'])){
			return $this->conf['apache_doc_root'];
		}

		$default = "";
		$this->apache_guess_from_conf();
		if(count($this->conf['webserver_docroot_arr']) == 0) {
			$possible_dirs = array(
					1 =>  '/usr/local/apache/htdocs',
					2 =>  '/usr/local/apache2/htdocs',
					3 =>  '/var/www/html',
					4 =>  '/var/lib/apache/htdocs',
					5 =>  '/var/lib/apache2/htdocs',
					6 =>  '/usr/local/httpd/htdoc',
					7 =>  '/home/httpd/html'
					);
			if(isset($this->conf['apache_root_dir'])) {
				array_unshift($possible_dirs, $this->make_path($this->conf['apache_root_dir'], "htdocs"));
			}
			$default = $this->search_dir($possible_dirs);
		}
		else if(count($this->conf['webserver_docroot_arr']) >1 && count($this->conf['webserver_docroot_arr']) <5) {
			array_push($this->conf['webserver_docroot_arr'], "other");

			while(true) {
				$num = $this->menubox("Please select the location of your Apache document root directory.\n".
						"This is the location where your web documents are stored.", $this->conf['webserver_docroot_arr']);

				if($this->conf['webserver_docroot_arr'][$num] == "other") {
					break;
				}
				else if($this->my_dir_exists($this->conf['webserver_docroot_arr'][$num])) {
					$this->conf['webserver_doc_root'] = $this->conf['webserver_docroot_arr'][$num];
					break;
				}

				$this->msgbox("Broken Apache configuration!\n".
						"Document root: \"".$this->conf['webserver_docroot_arr'][$num]."\" doesn't exist");
			}
		}
		else {
			$default = array_pop($this->conf['webserver_docroot_arr']);
		}

		return $default;
	}

	// APACHE_USER_GUESS
	function apache_user_guess()
	{
		if(!$this->is_using_apache()){
			return false;
		}

		if(!empty($this->conf['apache_user'])){
			return $this->conf['apache_user'];
		}

		$this->install_log("Trying to guess a user ID apache runs with ...");

		$this->apache_guess_from_conf();
		return empty($this->conf['apache_user'])? null : $this->conf['apache_user'];
	}

	// APACHE_ERRORLOG_GUESS
	function apache_errorlog_guess()
	{
		if(!empty($this->conf['apache_error_log'])){
			return $this->conf['apache_error_log'];
		}

		$this->apache_guess_from_conf();
		if(empty($this->conf['apache_error_log'])){
			$this->apache_guess_from_binary();
		}
	}

	// APACHE_GUESS_FROM_BINARY
	function apache_guess_from_binary()
	{
		if(!$this->is_using_apache()){
			return false;
		}

		$this->apache_exec_guess();

		$msg = "Guessed from Apache executable:";

		$httpd_res = $this->cmd2str($this->conf['apache_exec_file']. " -V 2>/dev/null");

		// Apache root
		if(!isset($this->conf['apache_root_dir']) &&
				preg_match("#HTTPD_ROOT=\"?([^\s\"]+)\"?\s*$#m", $httpd_res, $match)){
			$this->conf['apache_root_dir'] = $match[1];
			$this->install_log("$msg HTTPD_ROOT=".$this->conf['apache_root_dir']);
		}
		// Apache version
		if(empty($this->conf['apache_version']) &&
				preg_match("#^\s*Server version\:\s+Apache\/(\d)#m", $httpd_res, $match)){
			$this->conf['apache_version'] = $match[1];
			$this->install_log("$msg APACHE_VERSION=".$this->conf['apache_version']);
		}
		// Apache error log
		if(empty($this->conf['apache_error_log']) &&
				preg_match("#DEFAULT_ERRORLOG=\"?([^\s\"]+)\"?\s*$#m", $httpd_res, $match)){
			$this->conf['apache_error_log'] = $this->make_full_path($match[1], $this->conf['apache_root_dir']);
			$this->install_log("$msg ERROR_LOG=".$this->conf['apache_error_log']);
		}
		// Apache config file
		if(empty($this->conf['apache_conf_file']) &&
				preg_match("#SERVER_CONFIG_FILE=\"?([^\s\"]+)\"?\s*$#m", $httpd_res, $match)){
			$this->conf['apache_conf_file'] = $this->make_full_path($match[1], $this->conf['apache_root_dir']);
			$this->conf['apache_conf_dir'] = dirname($this->conf['apache_conf_file']);
			$this->install_log("$msg APACHE_CONFIG=".$this->conf['apache_conf_file']);
		}

		// guess Apache2 worker
		$httpd_l_res = $this->cmd2str($this->conf['apache_exec_file']. " -l 2>/dev/null");
		if(empty($this->conf['apache2_worker']) &&
				preg_match("#^\s*worker\.c#m",$httpd_l_res,$match)){
			$this->conf['apache2_worker'] = "yes";
			$this->install_log("$msg APACHE2 worker=".$this->conf['apache2_worker']);
		}

		return true;
	}

	// APACHE_GUESS_FROM_CONF
	function apache_guess_from_conf()
	{
		if(!$this->is_using_apache()){
			return false;
		}

		if(!isset($this->conf['apache_root_dir'])){
			$this->conf['apache_root_dir']="";
		}

		$this->apache_conf_read();

		$msg = "Guessed from Apache httpd.conf:";

		// Apache root
		if(empty($this->conf['apache_root_dir']) &&
				preg_match("#^\s*ServerRoot\s+\"?([^\s\"]+)\"?#m", $this->conf['apache_conf_cont'], $match)){
			$this->conf['apache_root_dir'] = $match[1];
			$this->install_log("$msg HTTPD_ROOT=".$this->conf['apache_root_dir']);
		}
		// Error log
		if(empty($this->conf['apache_error_log']) &&
				preg_match("#^\s*ErrorLog\s+\"?([^\s\"]+)\"?#m", $this->conf['apache_conf_cont'], $match)){
			$this->conf['apache_error_log'] = $this->make_full_path($match[1], $this->conf['apache_root_dir']);
			$this->install_log("$msg ERROR_LOG=".$this->conf['apache_error_log']);
		}
		// User
		if(empty($this->conf['apache_user']) &&
				preg_match("#^\s*User\s+([^\s]*)\s*$#m", $this->conf['apache_conf_cont'], $match)){
			$this->conf['apache_user'] = $match[1];
			$this->install_log("$msg USER=".$this->conf['apache_user']);
		}
		// Port
		if(empty($this->conf['apache_port'])){
			if(preg_match("#^\s*Port\s+([\d]+)\s*$#m", $this->conf['apache_conf_cont'], $match)){
				$port = $match[1];
			}
			else if(preg_match("#^\s*Listen\s+([\d]+)\s*$#m", $this->conf['apache_conf_cont'], $match)){
				$port = $match[1];
			}
			if(isset($port)){
				$this->conf['apache_port'] = $port;
				$this->install_log("$msg PORT=".$this->conf['apache_port']);
			}
		}
		// Group
		if(empty($this->conf['apache_group']) &&
				preg_match("#^\s*Group\s+([^\s]*)\s*$#m", $this->conf['apache_conf_cont'], $match)){
			$this->conf['apache_group'] = $match[1];
			$this->install_log("$msg GROUP=".$this->conf['apache_group']);
		}
		// MaxClients
		if(empty($this->conf['max_clients']) &&
				preg_match("@^\s*MaxClients\s+([\d]+)\s*$@m", $this->conf['apache_conf_cont'], $match)) {
			$this->conf['max_clients'] = $match[1];
			$this->install_log("$msg MaxClients=".$this->conf['max_clients']);
		}
		// Server Name
		if(empty($this->conf['server_name']) &&
			preg_match("@^\s*ServerName\s+([^\s]*).*<VirtualHost[^\>]+>.*</VirtualHost>@sU", $this->conf['apache_conf_cont'], $match)) {
		}

		// Virtual hosts
		if(empty($this->conf['apache_vhosts']) &&
				preg_match_all("@^\s*<VirtualHost([^\>]+)>(.*)</VirtualHost>@msU", $this->conf['apache_conf_cont'], $match)) {

			for($i=0; $i<count($match[0]); $i++) {

				if(preg_match("@\s+([^\:\s]+):?(\d*)@", $match[1][$i], $match2)) {

					$vhost_ip = $match2[1];
					$vhost_port = empty($match2[2]) ? "80" : $match2[2];

					if(preg_match("@^\s*DocumentRoot\s+\"?([^\s\"]+)\"?@m", $match[2][$i], $match2)) {
						$vhost_docroot = $match2[1];
					}
					if(preg_match("@^\s*ServerName\s+\"?([^\s\"]+)\"?@m", $match[2][$i], $match2)) {
						$vhost_servername = preg_replace("/:\d+/", "", $match2[1]);
					}

					if(isset($vhost_docroot) && $this->my_dir_exists($vhost_docroot) && isset($vhost_ip) && isset($vhost_servername) &&
						!preg_match("/\\\$/",$vhost_ip) && !preg_match("/\\\$/",$vhost_servername) )
					{

						$vhost_docroot = $this->nice_path($vhost_docroot);
						$this->conf['apache_vhosts'][$vhost_docroot.":".$vhost_ip] = array(
								'ip'     => $vhost_ip,
								'port'   => $vhost_port,
								'server' => $vhost_servername,
								'docroot'=> $vhost_docroot
								);
						$this->install_log("$msg VHOST=$vhost_ip:$vhost_port ($vhost_servername) => $vhost_docroot");
					}
				}
			}
		}
		// Document root
		if(empty($this->conf['webserver_docroot_arr']) &&
				preg_match_all("#^\s*DocumentRoot\s+\"?([^\s\"]+)\"?#m", $this->conf['apache_conf_cont'], $match)){

			$multiples = array();
			for($i=0, $c=1; $i<count($match[0]); $i++) {
				$docroot = $this->make_full_path($match[1][$i], $this->conf['apache_root_dir']);
				if( !strstr($docroot,'$') && $this->my_dir_exists($docroot) && !isset($multiples[
$docroot]) ) {
					$docroot = $this->nice_path($docroot);
					$this->conf['webserver_docroot_arr'][$c] = $docroot;
					$this->install_log("$msg DOCUMENT_ROOT=$docroot");
					$multiples[$docroot] = true;
					$c++;
				}
			}
		}
		// Libexec
		if(empty($this->conf['apache_libexec_dir'])){
			preg_match_all("#^\s*LoadModule\s+[^\s]+\s+\"?([^\s\"]+)\"?#m", $this->conf['apache_conf_cont'], $match);

			foreach($match[1] as $path){
				$path = $this->make_full_path($path, $this->conf['apache_root_dir']);
				if(preg_match("/libexec/", $path) || preg_match("/apache/", $path)){
					$this->conf['apache_libexec_dir'] = dirname($path);
					$this->install_log("$msg APACHE_LIBEXEC=".$this->conf['apache_libexec_dir']);
					break;
				}
			}
		}
		// PHP (module or executable) path
		if(empty($this->conf['php_path'])){
			// Script aliases
			preg_match_all("#^\s*ScriptAlias\s+\"?([^\s\"]+)\"?\s+\"?([^\s\"]+)\"?#m",
					$this->conf['apache_conf_cont'], $match);

			array_shift($match);

			$script_aliases = array();
			for($i=0; $i<count($match[0]); $i++){
				$script_aliases[$match[0][$i]] = $match[1][$i];
			}

			if(preg_match("#^\s*LoadModule\s+php\d_module\s+\"?([^\s\"]+)\"?\s*$#m",
						$this->conf['apache_conf_cont'], $match)){

				$this->conf['php_type'] = "module";
				$this->conf['php_path'] = $match[1];
			}
			elseif(preg_match("#^\s*Action\s+application\/x-httpd-php\s+\"?([^\s\"]+)\"?\s*$#m",
						$this->conf['apache_conf_cont'], $match)){
				$this->conf['php_type'] = "executable";
				$this->conf['php_path'] = $match[1];
			}
			if(!empty($this->conf['php_path'])){
				foreach($script_aliases as $alias => $path){
					$this->conf['php_path'] = preg_replace("#^$alias#", $path, $this->conf['php_path'], 1);
				}
				$this->conf['php_path'] = $this->make_full_path($this->conf['php_path'], $this->conf['apache_root_dir']);
				$this->install_log("$msg PHP_PATH=".$this->conf['php_path']);
				$this->install_log("$msg PHP_TYPE=".$this->conf['php_type']);
			}
		}

		return true;
	}

	// APACHE_HAS_MODULE
	function apache_has_module($mod_name)
	{
		$this->apache_exec_guess();

		if(preg_match("/$mod_name/", $this->cmd2str($this->conf['apache_exec_file']." -l"))) {
			return true;
		}
		return false;
	}

	// APACHE_GET_VERSION
	function apache_get_version()
	{
		if(!$this->is_using_apache()){
			return false;
		}

		if(!empty($this->conf['apache_version'])){
			return $this->conf['apache_version'];
		}

		$this->apache_guess_from_binary();

		if(empty($this->conf['apache_version'])){
			$version = $this->menubox(
					"The installation has failed to detect the version of Apache you use.".
					"Please choose the correct version.",
					array(1 => 'Apache 1.x.x', 2 => 'Apache 2.x.x')
					);
			$this->conf['apache_version'] = preg_replace("/Apache\s+(\d).*/", "\\1", $version);
		}

		return $this->conf['apache_version'];
	}

	// APACHE_DO_NEW_CONFIG
	function apache_do_new_config()
	{
		$apache_install_dir=dirname($this->conf['apache_root_dir']);
		if(empty($apache_install_dir) or $apache_install_dir == "/") {
			$this->on_error("Cannot install Apache into / (not permitted)");
		}
		if(!$this->my_dir_exists($apache_install_dir)){
			$this->on_error("Cannot install Apache into ".$apache_install_dir." (not a directory)");
		}

		$this->infobox("Configuring Apache installation ...");

		// Fix apachectl
		$this->conf['apachectl_path'] = $this->conf['apache_exec_dir']."/apachectl";
		$this->conf['apache_conf_file'] = $this->conf['apache_conf_dir']."/httpd.conf";

		$new_apachectl = preg_replace(
				"@(\s*HTTPD=).*$@m", "\\1\"".$this->conf['apache_exec_dir']."/httpd -f ".$this->conf['apache_conf_file']."\"",
				$this->file2str($this->conf['apachectl_path']));

		$new_apachectl = preg_replace(
				"@(\s*PIDFILE=).*$@m", "\\1".$this->conf['apache_root_dir']."/logs/httpd.pid", $new_apachectl);

		$this->my_fwrite($this->conf['apachectl_path'], "w", $new_apachectl);

		// Fix httpd.conf
		$this->conf['apache_conf_cont'] = preg_replace(
				"@/usr/local/apache@m", $this->conf['apache_root_dir'], $this->apache_conf_read());

		$this->apache_conf_close();

		$this->link_startup_daemon($this->conf['apachectl_path']);

		return true;
	}

	// APACHE_DO_CONFIG
	function apache_do_config($install_php, $install_webserver, $php_version)
	{
		$this->apache_conf_guess();
		$this->infobox("Configuring Apache ...");

		if($install_php){

			$this->apache_libexec_guess();

			if(!$this->my_dir_exists($this->conf['apache_libexec_dir'])) {
				$this->on_error("Broken Apache configuration: The needed directory:\n".
					$this->conf['apache_libexec_dir']."\nwhere Apache modules should be placed could not be found.\n".
					"The installation cannot proceed.\n");
			}

			$apache_libphp_so_path = $this->conf['apache_libexec_dir']."/libphp".$php_version.".so";

			if($this->my_file_exists($apache_libphp_so_path)){
				$this->my_backup($apache_libphp_so_path);
			}
			if(!$this->my_symlink($this->conf['prefix'].'/lib/libphp'.$php_version.'.so', $apache_libphp_so_path)){
				$this->on_error("Can't place a link to PHP library in your Apache ".
						"libexec directory at: ".$apache_libphp_so_path);
			}
		}

		/* backup httpd.conf */
		$apache_conf_backup = $this->my_backup($this->conf['apache_conf_file'], true);
		if(isset($apache_conf_backup)) {
			$this->msgbox("Backed-up Apache configuration file to:\n$apache_conf_backup");
		}

		$this->apache_conf_read();
		$config_exists = preg_match("/^\s*AddType\s+application.*php/m", $this->conf['apache_conf_cont']);

		/* delete old Zend configuration if any */
		$this->conf['apache_conf_cont'] = preg_replace(
				"@### Start of the.*### End of the ". $this->conf['product'] ." section.@s", "", $this->conf['apache_conf_cont']);

		/* If Use Existing PHP */
		if(!$install_php) {
			$apache_libphp_so_path="";
			if(!$config_exists){
				$this->msgbox("No existing PHP configuration found in your ".$this->conf['apache_conf_file']."!\n".
						"Your Apache must be configured to enable PHP scripts running on it,\n".
						"otherwise this product is useless!");
			}
		}
		else /* Install our own PHP conf */{
			$this->conf['apache_conf_cont'] = preg_replace(
				"@^\s*(AddType\s+application.*php)@m", "#Zend# \\1", $this->conf['apache_conf_cont']);

			$this->conf['apache_conf_cont'] = preg_replace(
							"@^\s*(Action\s+application/x-httpd-php)@m", "#Zend# \\1", $this->conf['apache_conf_cont']);

			$this->conf['apache_conf_cont'] = preg_replace(
							"@^\s*(AddModule\s+mod_php\d.c)@m", "#Zend# \\1", $this->conf['apache_conf_cont']);

			$this->conf['apache_conf_cont'] = preg_replace(
							"@^\s*(LoadModule\s+php\d_module)@m", "#Zend# \\1", $this->conf['apache_conf_cont']);

		}

		$this->apache_add_dir_defs($install_php, $apache_libphp_so_path, $php_version);

		// fix "Group" directive
		if($install_webserver && $this->my_file_exists($this->conf['groupfile'])){
			$g = file($this->conf['groupfile']);

			while(list( , $line) = each ($g)){
				$a = explode(":", $line);

				switch($a[0]){
					case "nobody":
						case "nogroup":
						$group = $a[0];
					break 2;
				}
			}
			if(isset($group)){
				$this->conf['apache_conf_cont'] =
					preg_replace ("@^\s*Group\s+.*@m", "Group $group", $this->conf['apache_conf_cont']);
			}
		}

		$this->apache_conf_close();
	}

	// APACHE_ADD_DIR_DEFS
	function apache_add_dir_defs($install_php, $apache_libphp_so_path, $php_version)
	{
		/* The following lines need to be added in any case */
		$new_section =
			"\n".
			"### Start of the ".$this->conf['product']." section.\n".
			"#\n".
			"# The configuration lines below are added by the ".$this->conf['product']." installation\n".
			"# script.\n".
			"#\n";

		if($install_php){
			$new_section .= "\nLoadModule php".$php_version."_module $apache_libphp_so_path\n\n";

			$new_section .=
				"<IfModule mod_mime.c>\n".
				"   AddType application/x-httpd-php .php\n".
				"   AddType application/x-httpd-php-source .phps\n".
				"</IfModule>\n\n";
		}

		$new_section .=
			"<Directory " .$this->conf['apache_doc_root']. "/ZendServerCenter\">\n".
			"   Options Indexes\n".
			"   DirectoryIndex index.html index.php\n".
			"</Directory>\n".
			"<Directory " . $this->conf['apache_doc_root'] . "/ZendInformationCenter\">\n".
			"   Options Indexes\n".
			"   DirectoryIndex index.html index.php\n".
			"</Directory>\n\n".
			"### End of the ". $this->conf['product'] ." section.\n";

		$this->conf['apache_conf_cont'] = $new_section."\n\n".$this->conf['apache_conf_cont'];
	}

	// APACHE_STOP
	function apache_stop()
	{
		if(!$this->is_using_apache()){
			return false;
		}

		$this->apache_exec_guess();
		if(!empty($this->conf['apachectl_path'])) {
			if(!strstr($this->file2str($this->conf['apachectl_path']), "startssl")) {

				$this->progress_bar_start(20, "Stopping Apache ...\n".
				"Please wait... (this may take a few seconds)", "Stopping ...");

				$status = $this->run_command($this->conf['apachectl_path']." stop 1>/dev/null");
				sleep(4);

				$this->progress_bar_stop();
				if($status==0){
					return true;
				}
			}
			else {
				$this->install_log("Refusing to stop Apache SSL");
			}
		}
		else{
			$this->install_log("Apache control utility (apachectl) not found");
		}
		return false;
	}

	// APACHE_START
	function apache_start()
	{
		if(!$this->is_using_apache()){
			return false;
		}

		$this->apache_exec_guess();
		if(!empty($this->conf['apachectl_path'])) {
			if(!strstr($this->file2str($this->conf['apachectl_path']), "startssl")) {

				$this->progress_bar_start(20, "Starting Apache ...\n".
				"Please wait... (this may take a few seconds)", "Starting ...");

				$status = $this->run_command($this->conf['apachectl_path']." start 1>/dev/null");
				sleep(4);

				$this->progress_bar_stop();
				if($status==0){
					return true;
				}
			}
			else {
				$this->install_log("Refusing to start Apache SSL");
			}
		}
		else{
			$this->install_log("Apache control utility (apachectl) not found");
		}
		return false;
	}

	// APACHE_RESTART
	function apache_restart()
	{
		if(!$this->is_using_apache()){
			return false;
		}

		$this->apache_exec_guess();
		if(!empty($this->conf['apachectl_path'])) {

			$stop_cmd = "stop";
			$start_cmd = "start";

			$this->progress_bar_start(20, "Restarting Apache ...\n".
			"Please wait... (this may take a few seconds)", "Restarting ...");

			if(strstr($this->file2str($this->conf['apachectl_path']), "startssl")) {
				$status = $this->run_command($this->conf['apachectl_path']." graceful 1>/dev/null");
				sleep(4);
			}
			else {
				$this->run_command($this->conf['apachectl_path']." $stop_cmd 1>/dev/null");
				sleep(4);
				$status = $this->run_command($this->conf['apachectl_path']." $start_cmd 1>/dev/null");
				sleep(4);
			}

			$this->progress_bar_stop();

			if($status==0){
				return true;
			}
		}
		else{
			$this->install_log("Apache control utility (apachectl) not found");
		}
		return false;
	}



	// ----------------------
	// ZEUS FUNCTIONS
	// ======================

	// ZEUS_RESTART
	function zeus_restart()
	{
		$this->webserver_root_guess();
		$this->infobox("Restarting Zeus ...");

		if($this->my_file_exists($this->conf['zeus_root']."/restart-zeus")){
			$status = $this->run_command($this->conf['zeus_root']."/restart-zeus");
			if($status == 0){
				return true;
			}
		}
		return false;
	}

	// ZEUS_DOCROOT_GUESS
	function zeus_docroot_guess()
	{
		if(!empty($this->conf['zeus_doc_root'])){
			return $this->conf['zeus_doc_root'];
		}

		$this->webserver_root_guess();

		$multiples = array();
		$c = 1;
		if(!empty($this->conf['zeus_root'])){
			if($this->my_dir_exists($this->conf['zeus_root']."/webadmin/conf/virtual_servers/sites")) {
				foreach ($this->my_list_dir($this->conf['zeus_root']."/webadmin/conf/virtual_servers/sites", "*") as $file) {
					if(preg_match("/^docroot\s+(.*)$/m", $this->file2str($file), $match)) {
						if($this->my_dir_exists($match[1]) && !isset($multiples[$match[1]])) {
							$this->conf['webserver_docroot_arr'][$c] = $match[1];
							$c++;
						}
					}
				}
			}
		}

		$default = "";
		if(count($this->conf['webserver_docroot_arr']) == 0) {
			$default = $this->search_dir(array(
						1	=>	"/usr/local/zeus/web/docroot",
						2	=>	"/var/www/html"
						));
		}
		else if(count($this->conf['webserver_docroot_arr']) > 1) {
			array_push($this->conf['webserver_docroot_arr'], "other");
			while(true) {
				$num = $this->menubox("Please select the location of your document root directory.\n".
						"This is the location where your web documents are stored.", $this->conf['webserver_docroot_arr']);

				if($this->conf['webserver_docroot_arr'][$num] == "other") {
					break;
				}
				else if($this->my_dir_exists($this->conf['webserver_docroot_arr'][$num])) {
					$this->conf['webserver_doc_root'] = $this->conf['webserver_docroot_arr'][$num];
					break;
				}

				$this->msgbox("Broken Zeus configuration!\n".
						"Document root: \"".$this->conf['webserver_docroot_arr'][$num]."\" doesn't exist");
			}
		}
		else {
			$default = array_pop($this->conf['webserver_docroot_arr']);
		}

		return $default;
	}

	// ZEUS_USER_GUESS
	function zeus_user_guess()
	{
		if(isset($this->conf['zeus_user'])) {
			return $this->conf['zeus_user'];
		}

		$this->webserver_root_guess();

		if(!empty($this->conf['zeus_root'])){
			if($this->my_file_exists($this->conf['zeus_root']."/web/global.cfg")) {
				if(preg_match("/^uid\s+(.*)$/m", $this->file2str($this->conf['zeus_root']."/web/global.cfg"), $match)) {
					$this->conf['zeus_user'] = $match[1];
				}
			}
		}

		return isset($this->conf['zeus_user']) ? $this->conf['zeus_user'] : null;
	}



	// ----------------------
	// SUNONE FUNCTIONS
	// ======================

	// SUNONE_RESTART
	function sunone_restart()
	{
		$this->sunone_root_guess();
		$this->infobox("Restarting SunONE ...");

		if($this->my_file_exists($this->conf['sunone_root']."/restart")){
			$status = $this->run_command($this->conf['sunone_root']."/restart");
			if($status == 0){
				return true;
			}
		}
		return false;
	}

	// SUNONE_ROOT_GUESS
	function sunone_root_guess()
	{
		if(!empty($this->conf['sunone_root'])) {
			return $this->conf['sunone_root'];
		}

		$this->webserver_root_guess();

		$c = 1;
		$servers = array();
		foreach ($this->my_list_dir($this->conf['sunone_root'], "https-*") as $dir) {
			$server_id = substr($dir, strrpos($dir, "-")+1);
			if($server_id == "admserv") {
				continue;
			}
			$servers[$c++] = $server_id;
		}

		if(count($servers) > 1) {
			$i = $this->menubox("Please select the server instance you wish to use.", $servers);
			$server_id = $servers[$i];
		}
		else $server_id = $servers[1];

		$this->conf['sunone_root'] = $this->make_path($this->conf['sunone_root'], "https-".$server_id);
	}

	// SUNONE_GUESS_FROM_CONF
	function sunone_guess_from_conf()
	{
		$this->sunone_root_guess();

		$sunone_conf = $this->conf['sunone_root']."/config/server.xml";
		if(!$this->my_file_exists($sunone_conf)) {
			$this->install_log("Can't find SunONE configuration file: $sunone_conf");
			return false;
		}
		$sunone_conf_cont = $this->file2str($sunone_conf);

		$msg = "Guessed from SunONE configuration:";

		// User
		if(empty($this->conf['sunone_user']) &&
				preg_match("#PROPERTY\s+name=\"user\"\s+value=\"([^\"]*)\"#mi", $sunone_conf_cont, $match)){
			$this->conf['sunone_user'] = $match[1];
			$this->install_log("$msg USER=".$this->conf['sunone_user']);
		}
		// Port
		if(empty($this->conf['sunone_port']) &&
				preg_match("#port=\"(\d+)\"#mi", $sunone_conf_cont, $match)){
			$this->conf['sunone_port'] = $match[1];
			$this->install_log("$msg PORT=".$this->conf['sunone_port']);
		}
		// Group
		if(empty($this->conf['sunone_group']) &&
				preg_match("#PROPERTY\s+name=\"group\"\s+value=\"([^\"]*)\"#mi", $sunone_conf_cont, $match)){
			$this->conf['sunone_group'] = $match[1];
			$this->install_log("$msg GROUP=".$this->conf['sunone_group']);
		}
		// Server Name
		if(empty($this->conf['server_name']) &&
			preg_match("#servername=\"[^\"]*\"#mi", $sunone_conf_cont, $match)){
		}
		// Document root
		if(empty($this->conf['webserver_docroot_arr']) &&
				preg_match_all("#PROPERTY\s+name=\"docroot\"\s+value=\"([^\"]*)\"#mi", $sunone_conf_cont, $match)){

			$multiples = array();
			for($i=0, $c=1; $i<count($match[0]); $i++) {
				$docroot = $match[1][$i];
				if(!strstr($docroot,'$') && $this->my_dir_exists($docroot) && !isset($multiples[$docroot])) {
					$this->conf['webserver_docroot_arr'][$c] = $docroot;
					$this->install_log("$msg DOCUMENT_ROOT=$docroot");
					$multiples[$docroot] = true;
					$c++;
				}
			}
		}

		return true;
	}

	// SUNONE_DOCROOT_GUESS
	function sunone_docroot_guess()
	{
		if(!empty($this->conf['sunone_doc_root'])){
			return $this->conf['sunone_doc_root'];
		}

		$this->sunone_guess_from_conf();

		if(count($this->conf['webserver_docroot_arr']) > 1) {
			array_push($this->conf['webserver_docroot_arr'], "other");
			while(true) {
				$num = $this->menubox("Please select the location of your document root directory.\n".
						"This is the location where your web documents are stored.", $this->conf['webserver_docroot_arr']);

				if($this->conf['webserver_docroot_arr'][$num] == "other") {
					break;
				}
				else if($this->my_dir_exists($this->conf['webserver_docroot_arr'][$num])) {
					$this->conf['webserver_doc_root'] = $this->conf['webserver_docroot_arr'][$num];
					break;
				}

				$this->msgbox("Broken SunONE configuration!\n".
						"Document root: \"".$this->conf['webserver_docroot_arr'][$num]."\" doesn't exist");
			}
		}
		else {
			$default = array_pop($this->conf['webserver_docroot_arr']);
		}

		return $default;
	}

	// SUNONE_USER_GUESS
	function sunone_user_guess()
	{
		if(!empty($this->conf['sunone_user'])) {
			return $this->conf['sunone_user'];
		}

		$this->sunone_guess_from_conf();

		if(!empty($this->conf['sunone_root'])) {
			$conf_file = $this->conf['sunone_root']."/config/magnus.conf";
			if($this->my_file_exists($conf_file)) {
				if(preg_match("/^\s*User\s+(.*)$/m", $this->file2str($conf_file), $match)) {
					$this->conf['sunone_user'] = $match[1];
				}
			}
		}

		return (isset($this->conf['sunone_user']) ? $this->conf['sunone_user'] : null);
	}



	// ----------------------
	// PHP FUNCTIONS
	// ======================

	// DONT_HAVE_PHP
	function dont_have_php()
	{
		$this->on_error("In order to install the ".$this->conf['product'].", you must have supported PHP\n".
				"installed on your computer. Please install a valid PHP,\nthen restart the installation.");
	}

	// PHP_VERSION_DETECT
	function php_version_detect($silent = false, $warn_unsupported_php = false)
	{
		if(!$this->php_version_detect_from_apache($silent)){
			if(!$silent) {
				if(!$this->php_version_ask()){
					$this->php_version_select_manually();
				}
			}
		}

		if(isset($this->conf['php_version'])) {
			$this->install_log("Detected PHP version: ".$this->conf['php_version']);

			if($this->conf['php_version'] == "unsupported" && $warn_unsupported_php) {
				$this->msgbox(
					"Installation has detected PHP version ".$this->conf['detected_unsupported_php_version'].".\n".
					"This version is not compatible with the ".$this->conf['product']." ".$this->conf['version'].".\n".
					"The installation process will continue but in order for the ".$this->conf['product']."\n".
					"to load you must switch to a compatible PHP version.\n\n".
					"For a detailed compatibility table Please refer to\n".
					"http://www.zend.com/store/products/product_compatibility.php");
			}
		}
	}

	// PHP_VERSION_DETECT_FROM_APACHE
	function php_version_detect_from_apache($silent = false)
	{
		$this->install_log("Detecting PHP version ...");

		if($this->is_using_apache($silent)){

			$this->apache_conf_guess($silent);
			$this->php_path_guess($silent);

			if(isset($this->conf['php_type'])) {
				if($this->conf['php_type'] == "unknown"){
					$this->apache_exec_guess($silent);
					if($this->php_version_search_in($this->conf['apache_exec_file'])){
						$this->conf['php_type'] = "builtin";
					}
				}
				elseif($this->conf['php_type'] == "executable" && isset($this->conf['php_path'])){
					if(preg_match("/(\d\.\d\.\d)/", $this->cmd2str($this->conf['php_path']. " -v 2>/dev/null"), $match)){
						$this->conf['php_version'] = $match[1];
					}
				}
				elseif($this->conf['php_type'] == "module" && isset($this->conf['php_path'])){
					$this->php_version_search_in($this->conf['php_path']);
				}
			}
		}

		if(isset($this->conf['php_version'])){
			if($silent) {
				if(!$this->php_version_substitute($silent)) {
					$this->conf['php_version'] = "unsupported";
					return false;
				}
				else return true;
			}
			return $this->php_version_confirm();
		}

		return false;
	}

	// PHP_VERSION_ASK
	function php_version_ask($msg=null, $type=null)
	{
		$this->php_versions_array_init();
		
		if(!$this->conf['interactive']) {
			return false;
		}	

		if($msg == null){
			$msg = "Setup has failed to determine your PHP version.";
		}

		$menu = array(
				1 => "Detection of PHP version from a binary.",
				2 => "Manually select my PHP version."
				);
		$choice = $this->menubox($msg, $menu);

		if($choice == 1){
			return $this->php_version_detect_from_binary();
		}
		elseif($choice == 2){
			$this->php_version_select_manually();
			return true;
		}
	}

	// PHP_VERSION_DETECT_FROM_BINARY
	function php_version_detect_from_binary()
	{
		$this->php_versions_array_init();

		$this->php_type_guess();

		$confirm_msg = "Setup will attempt to determine your PHP version from a binary.\n";

		if($this->conf['php_type'] == "module") // detect PHP version from a PHP module
		{
			$default = empty($this->conf['php_path'])? "" : $this->conf['php_path'];
			$this->conf['php_path'] = $this->select_file(
					$confirm_msg."Please enter the full path to your PHP module", $default);

			$this->php_version_search_in($this->conf['php_path']);
		}
		elseif($this->conf['php_type'] == "executable") // detect from a PHP binary
		{
			$default = empty($this->conf['php_path'])? "" : $this->conf['php_path'];
			$this->conf['php_path'] = $this->select_file(
					$confirm_msg."Please enter the full path to your PHP executable file", $default);

			if(preg_match("/(\d\.\d\.\d)/", $this->cmd2str($this->conf['php_path']. " -v 2>/dev/null"), $match)){
				$this->conf['php_version'] = $match[1];
			}
		}
		elseif($this->conf['php_type'] == "builtin") // detect from an Apache executable
		{
			$default = empty($this->conf['apache_exec_file'])? "": $this->conf['apache_exec_file'];
			$php_search_path = $this->select_file(
					$confirm_msg."Please enter the full path to your ".$this->conf['webserver'], $default);

			$this->php_version_search_in($php_search_path);
		}
		elseif($this->conf['php_type'] == "fastcgi") // detect from a PHP FastCGI SAPI executable
		{
			$php_search_path = $this->select_file(
					$confirm_msg."Please enter the full path to your FastCGI PHP application", "");

			$this->php_version_search_in($php_search_path);
		}

		if(empty($this->conf['php_version'])){
			$this->msgbox("The installation still fails to determine your PHP version.");
			return false;
		}

		return $this->php_version_confirm();
	}

	// PHP_VERSION_SELECT_MANUALLY
	function php_version_select_manually($msg="")
	{
		$this->php_versions_array_init();

		if(!$this->conf['interactive']) {
			return false;
		}

		foreach ($this->conf['php_versions_array'] as $num => $php_arr){
			if($php_arr[2] != 0){
					$sub_str=substr($php_arr[0],0 ,(strlen($php_arr[0])-1) );
					$php_ver_arr[$num]= $php_arr[0]." (".$sub_str.$php_arr[2]." and later)";
			}else{
				$php_ver_arr[$num]= $php_arr[0];
			}
			$last_i = $num+1;
		}

		$msg = "$msg\nIn order to install the correct version of the Zend Modules\n".
			"you must specify which PHP version you are using.";

		$php_ver_arr[$last_i] = "I don't have PHP";
		$choice = $this->menubox($msg , $php_ver_arr);

		if($choice == $last_i) {
			if(isset($this->conf['doesnt_need_php'])) {
				$this->conf['php_version'] = "donthave";
			}
			else $this->dont_have_php();
		}
		else $this->conf['php_version'] = $this->conf['php_versions_array'][$choice][1];

		return true;
	}

	// PHP_TYPES_SET
	function php_types_set($php_types_arr)
	{
		$this->install_log("PHP types was set to: ". var_export($php_types_arr, true));
		$this->conf['php_types_supported'] = $php_types_arr;
	}

	// PHP_TYPE_ADD
	function php_type_add($new_php_type) // PHP type is an array that consists of two elements:
	                                     // 1) PHP type description
																			 // 2) PHP type internal use name
	{
		$this->install_log("Added new PHP type: ". $new_php_type[0]);
		array_push($this->conf['php_types_supported'], $new_php_type);
	}

	// PHP_TYPE_EXISTS
	function php_type_exists($php_type)
	{
		$this->php_types_array_init();

		if(!empty($this->conf['php_types_supported'])){
			foreach($this->conf['php_types_supported'] as $php_type_arr){
				if($php_type_arr[1] == $php_type){
					return true;
				}
			}
		}
		return false;
	}

	// PHP_TYPE_GUESS
	function php_type_guess()
	{
		$this->install_log("Trying to guess PHP type ...");

		$this->php_types_array_init();
		$this->apache_guess_from_conf();

		if(!empty($this->conf['php_type']) && $this->conf['php_type'] != "unknown"){
			$this->php_type_supported_check();
			return true;
		}
		else {
			if($this->is_using_apache() && $this->apache_has_module("mod_php")) {
				$this->conf['php_type'] = "builtin";
				$this->php_type_supported_check();
				return true;
			}
		}

		$old_php_type = isset($this->conf['php_type'])? $this->conf['php_type'] : "unknown";

		$this->install_log("Old PHP type was: $old_php_type");

		while(1) {
			if(count($this->conf['php_types_supported'])==1){
				if($this->yesnobox("Are you using ".$this->conf['php_types_supported'][1][0]."?")){
					$this->conf['php_type'] = $this->conf['php_types_supported'][1][1];
				}
				else{
					$this->on_abort("For now ".$this->conf['product']." supports only ".
							$this->conf['php_types_supported'][1][0]." with ".$this->conf['webserver'], true);
				}
			}
			else {
				foreach($this->conf['php_types_supported'] as $num => $data){
					$php_types[$num] = $data[0];
					$last_i = $num+1;
				}

				$php_types[$last_i] = "I don't have PHP";
				$choice = $this->menubox("Please choose the type of PHP you use", $php_types, false,
						"The ".$this->conf['product']." supports the following PHP configurations:\n\n".
						$this->array2string($this->conf['php_types_supported'], 0));

				if($choice == $last_i) {
					if(isset($this->conf['doesnt_need_php'])) {
						$this->conf['php_type'] = "donthave";
					}
					else $this->dont_have_php();
				}
				else $this->conf['php_type'] = $this->conf['php_types_supported'][$choice][1];
			}

			if($this->conf['php_type'] != "executable") { /* Dummy check */
				if(isset($this->conf['php_path'])) {
					if(preg_match("/^([\d\.]+)$/", $this->cmd2str($this->conf['php_path']))) {
						$this->msgbox("You have specified that you are using PHP ".$this->conf['php_type']."\n".
								"But the installation detected usage of PHP CGI.\nPlease don't try to cheat the installation.");
						continue;
					}
				}
			}
			break;
		}
		$this->install_log("Guessed PHP type: ".$this->conf['php_type']);

		if($this->conf['php_type'] != $old_php_type){
			unset($this->conf['php_path']);
		}

		return true;
	}

	// PHP_TYPE_SUPPORTED_CHECK
	function php_type_supported_check()
	{
		$this->install_log("Checking eigther PHP type (".$this->conf['php_type'].") is supported.");

		$this->php_types_array_init();

		if(empty($this->conf['php_type']) || $this->conf['php_type'] == "unknown") {
			$this->php_type_guess();
		}

		if(!$this->php_type_exists($this->conf['php_type'])) {

			$this->on_abort("ERROR: Unsupported PHP configuration.\n".
					"The ".$this->conf['product']." supports the following PHP configurations:\n\n".
					$this->array2string($this->conf['php_types_supported'], 0), true);
		}
	}

	// PHP_VERSIONS_ARRAY_INIT
	function php_versions_array_init()
	{
		if(@isset($this->conf['php_versions_array_initialized'])){
			return;
		}
		$this->install_log("Initializing array of PHP versions");
		$this->conf['php_versions_array'] = array(
				1 =>  array("PHP 4.0.5/6", "4.0.6", 0),
				2 =>  array("PHP 4.1.x", "4.1.x", 0),
				3 =>  array("PHP 4.2.0", "4.2.0", 0),
				4 =>  array("PHP 4.2.x (other than 4.2.0)", "4.2.x", 0),
				5 =>  array("PHP 4.3.x", "4.3.x", 0),
				6	=>	array("PHP 5.0.0", "5.0.0", 0)
				);
		$this->conf['php_versions_array_initialized'] = true;
	}

	// PHP_TYPES_ARRAY_INIT
	function php_types_array_init()
	{
		if(!empty($this->conf['php_types_supported'])){
			return;
		}
		$this->conf['php_types_supported'] = array(
				1 => array("PHP module (e.g libphp4.so)", "module"),
				2 => array("PHP executable (also called CGI)", "executable"),
				3 => array("PHP compiled into ". $this->conf['webserver'], "builtin"),
				4 => array("Other", "unknown")
				);
	}

	// PHP_VERSION_SUBSTITUTE
	function php_version_substitute($silent = false)
	{
		$this->php_versions_array_init();

		if(!isset($this->conf['php_version'])){
			if($silent) {
				return false;
			}
			$this->php_version_select_manually();
		}

		foreach ($this->conf['php_versions_array'] as $num => $php_arr){ /* look for an exact match */
			if($php_arr[1] == $this->conf['php_version']){
				return true;
			}
			elseif (($php_arr[1] == "4.0.6" ) && ("4.0.5" == $this->conf['php_version'])){
				$this->conf['php_version'] = $php_arr[1];
				return true;
			}
		}
		foreach ($this->conf['php_versions_array'] as $num => $php_arr){
			if($this->ver_equal($php_arr[1], $this->conf['php_version'], $php_arr[2])){ /* look for a corresponding match */
				$this->conf['php_version'] = $php_arr[1];
				return true;
			}
		}

		$this->conf['detected_unsupported_php_version'] = $this->conf['php_version'];
		return false;
	}

	// PHP_VERSION_CONFIRM
	function php_version_confirm()
	{
		$this->php_versions_array_init();

		$this->php_type_supported_check();

		if($this->yesnobox("Please confirm the PHP version you use: ".$this->conf['php_version'])){
			if($this->php_version_substitute()){
				$this->install_log("Confirmed PHP version: ".$this->conf['php_version']);
				return true;
			}

			if(count($this->conf['php_versions_array']) != 1) {
				$vers_msg = "on of the following versions of PHP:\n\n"
					.$this->array2php_support_string($this->conf['php_versions_array'], 0, 2);
			}
			else {
				$vers_msg = $this->array2php_support_string($this->conf['php_versions_array'], 0, 2);
			}

			$this->on_error($this->conf['product']." does not support PHP ".$this->conf['php_version'].
					"\nIn order to install ".$this->conf['product']." on ".$this->conf['uname']['sysname']."\n".
					"you must have $vers_msg");
		}

		unset($this->conf['php_version']);
		return false;
	}

	// PHP_VERSIONS_GET_ARRAY
	function php_versions_get_array()
	{
		$this->php_versions_array_init();

		$php_vers_array = array();
		foreach($this->conf['php_versions_array'] as $php_ver) {
			array_push($php_vers_array, $php_ver[1]);
		}

		return $php_vers_array;
	}

	// PHP_VERSION_ADD
	function php_version_add($descr, $ver, $from=0) /* description of version , PHP version */
	{
		$this->php_versions_array_init();

		$descr = trim($descr);
		$ver = trim($ver);

		$found = false;
		foreach($this->conf['php_versions_array'] as $num => $php_arr) {
			if($php_arr[1] == $ver) {
				$found = true;
				$this->conf['php_versions_array'][$num] = array($descr, $ver, $from);
				break;
			}
		}
		if(!$found) {
			if(count($this->conf['php_versions_array']) == 0) {
				// Always begin the index from 1
				$this->conf['php_versions_array'][1] = array($descr, $ver, $from);
			}
			else array_push($this->conf['php_versions_array'], array($descr, $ver, $from));
		}

		$this->install_log("Added $ver ($descr) from version $from to the array of PHP versions");
	}

	// PHP_VERSION_REMOVE
	function php_version_remove($ver)
	{
		$this->php_versions_array_init();

		$new_array = array();
		$index = 1;

		foreach($this->conf['php_versions_array'] as $num => $php_arr){
			if($php_arr[1] != $ver){
				$new_array[$index++] = $php_arr;
			}
			else{
				$this->install_log("Removed $ver from array of PHP versions");
			}
		}

		$this->conf['php_versions_array'] = $new_array;
	}

	// PHP_VERSIONS_SET
	function php_versions_set($php_arr)
	{
		$this->install_log("PHP versions set as: ". var_export($php_arr, true));
		$this->conf['php_versions_array'] = $php_arr;
	}

	// PHP_VERSION_SEARCH_IN
	function php_version_search_in($file)
	{
		$this->progress_bar_start(100, "Trying to determine the version of PHP.\n".
				"Please wait... (this may take a few seconds)", "Detecting ...");

		$ret = false;
		if(!$this->my_file_exists($file)){
			$this->install_log("Path to PHP: $file is broken");
		}
		else if($this->is_broken_link($file)) {
			$this->warn("Link: $file is broken");
		}
		else {
			$file_cont = @file_get_contents($file);
			if(preg_match("/PHP\/(\d\.\d\.\d)/", $file_cont, $match)){
				$this->conf['php_version'] = $match[1];
				$ret = $match[1];
			}
		}

		$this->progress_bar_stop();
		return $ret;
	}

	// PHP_PATH_GUESS
	function php_path_guess()
	{
		if(empty($this->conf['php_type'])){
			$this->conf['php_type'] = "unknown";
		}
		if(!empty($this->conf['apache_conf_file'])){
			$this->apache_guess_from_conf();
		}
	}

	// PHP_INI_LOCATION_GUESS
	function php_ini_location_guess()
	{
		if(isset($this->conf['php_ini_path'])) {
			return $this->conf['php_ini_path'];
		}

		$this->install_log("Trying to guess PHP.ini location ...");

		$default = array(
				1 =>  "/usr/local/lib",
				2 =>  $this->conf['prefix']."/etc",
				3 =>  "/usr/local/Zend/etc",
				4 =>  "/usr/local/etc",
				5 =>  "/etc",
				6	=>	"/etc/php"
				);

		foreach ($default as $path){
			if($this->my_file_exists("$path/php.ini")){
				$this->install_log("Found php.ini in: $path");
				$php_ini_dir = $path;
				break;
			}
		}

		if(empty($php_ini_dir)){
			$php_ini_dir = "";
		}

		if(isset($this->conf['default_php_ini_path'])) {
			$php_ini_dir = $this->conf['default_php_ini_path'];
		}

		while(true){
			$default_path = $php_ini_dir;
			$word = ($default_path == "") ? "Enter" : "Confirm";
			$php_ini_dir = $this->select_dir("$word the location of your php.ini file", $default_path);
			if(!$this->my_file_exists("$php_ini_dir/php.ini")){
				$this->msgbox("php.ini was not found in $php_ini_dir");
			}
			else{
				$this->conf['php_ini_path'] = "$php_ini_dir/php.ini";
				$this->strip_symlink($this->conf['php_ini_path']);
				break;
			}
		}

		$this->conf['saved_config']['default_php_ini_path'] = $php_ini_dir;
		$this->install_log("Confirmed the location of php.ini at: $php_ini_dir");
	}

	// PHP_INI_FIX
	function php_ini_fix()
	{
		$this->install_log("Trying to fix PHP.ini ...");
		$this->php_ini_cont_check();
		$this->php_ini_replace_entry("\' at MYSQL_PORT", "; at MYSQL_PORT");
		$this->install_log("PHP.ini fixed");
	}

	// GUESS_CONFLICT
	function guess_conflict()
	{
		$this->install_log("Trying to guess a conflict in PHP.ini ...");

		$not_asked=TRUE;

		foreach ($this->conf['php_ini_cont'] as $num => $line){
			$line = chop($line);
			if(preg_match("#^\s*zend_extension(?:_ts)?=.*ZendAccelerator.*#", $line) ||
				preg_match("#^\s*zend_extension(?:_ts)?=.*ZendPerformanceSuite.*#", $line)){

				if($not_asked){
					$not_asked=FALSE;
					$config_choose = $this->menubox("You appear to have an old version of the Zend Performance Suite (ZPS)\n".
													"installed, that is incompatible with this version of the Zend Studio\n".
													"Server.  If you choose to continue the installation, ZPS will be\n".
													"disabled.  To re-enable it  please download and install the Zend\n".
													"Performance Suite version 3.5.0a or later.",

								array(  1    =>  "Disable ZPS and continue",
										2   =>  "Abort installation"));
				}

				if($config_choose == 1){
					$this->conf['php_ini_cont'][$num] = preg_replace(
						"@^\s*(.*)@", ";Zend; \\1", $this->conf['php_ini_cont'][$num]);
						$this->install_log("Commenting PHP.ini entry: $line");
				}else{
					$this->default_abort();
				}
			}
		}

		$this->install_log("PHP.ini checked for a conflict");
	}

	// PHP_INI_REORDER
	function php_ini_reorder()
	{
		$this->install_log("Trying to re-order PHP.ini ...");

		$this->php_ini_cont_check();

		$zend_ext_order_table = array(
				"ZendExtensionManager"    => "",
				"ZendExtensionManager_TS" => "",
				"ZendOptimizer"		        => "",
				"ZendOptimizer_TS"        => "",
				"ZendAccelerator"         => "",
				"ZendDebugger"            => "",
				"ZendDebugger_TS"         => ""
				);

		foreach($this->conf['php_ini_cont'] as $line_num => $line){
			$line = chop($line);
			if(preg_match("/^\s*zend_extension(?:_ts)?=([a-zA-Z0-9_\/\"-\.]*Zend[a-zA-Z0-9_]+\.[a-z]+)/",
						$line, $match)){
				$zend_module = basename($match[1]);
				$zend_module = preg_replace("/\.[^\.]+$/", "", $zend_module);
				$zend_ext_order_table[$zend_module] = $line;
				$this->conf['php_ini_cont'][$line_num] = "";
			}
		}

		foreach($zend_ext_order_table as $module => $line){
			if(strlen($line) > 0) {
				array_push($this->conf['php_ini_cont'], "$line\n");
				$this->conf['php_ini_modified'] = true;
			}
		}

		$this->install_log("PHP.ini re-ordered");
	}

	// PHP_INI_REMOVE_ENTRY
	function php_ini_remove_entry($value)
	{
		$this->php_ini_cont_check();

		reset($this->conf['php_ini_cont']);
		foreach ($this->conf['php_ini_cont'] as $num => $line){

			if(!preg_match("/^\s*;/", $line) && preg_match("#$value#", $line)){
				$this->conf['php_ini_cont'][$num] = preg_replace(
				"@^\s*(.*$value)@m", ";Zend; \\1", $this->conf['php_ini_cont'][$num]);
				$this->install_log("Commenting PHP.ini entry: $line");
			}
		}
	}

	// IN_PHP_INI
	function in_php_ini($key)
	{
		$this->php_ini_cont_check();
		reset($this->conf['php_ini_cont']);
		foreach ($this->conf['php_ini_cont'] as $num => $line){

			$line = preg_replace("/;.*$/", "", $line);

			if(preg_match("/\s*$key\s*=/", $line)){
				return true;
			}
		}
		return false;
	}

	// PHP_INI_GET_ENTRY
	function php_ini_get_entry($key)
	{
		$this->php_ini_cont_check();
		reset($this->conf['php_ini_cont']);
		foreach ($this->conf['php_ini_cont'] as $num => $line){

			$line = preg_replace("/;.*$/", "", $line);

			if(preg_match("/\s*$key\s*=(.*)/", $line, $match)){
				return trim($match[1], "\"\' \t");
			}
		}
		return "";
	}

	// PHP_INI_ADD_ENTRY
	function php_ini_add_entry($key, $value, $replace_old_entries=true)
	{
		$this->php_ini_cont_check();

		$new_line = "$key=$value". $this->conf['nl'];
		reset($this->conf['php_ini_cont']);
		$replaced = false;

		foreach ($this->conf['php_ini_cont'] as $num => $line){

			$line = preg_replace("/;.*$/", "", $line);

			if(preg_match("/\s*$key\s*=/", $line)){
				$replaced = true;

				if($replace_old_entries) {
					if($key == "session.save_path"){
						preg_match("/^\s*$key\s*=\s*\"?([^\"]*)\"?\s*$/", $line , $match);
						if(!$this->my_dir_exists($match[1])){
							$this->conf['php_ini_cont'][$num] = $new_line;
						}
					}
					else{
						$this->conf['php_ini_cont'][$num] = $new_line;
					}

					$this->install_log("Replaced line: ".$this->conf['php_ini_cont'][$num]." by: $new_line in php.ini");
				}
			}
		}

		if(!$replaced){
			array_push($this->conf['php_ini_cont'], $new_line);
			$this->install_log("Added new line to php.ini: $new_line");
		}

		$this->conf['php_ini_modified'] = true;
	}

	// PHP_INI_REPLACE_ENTRY
	function php_ini_replace_entry($old, $new)
	{
		$this->php_ini_cont_check();

		reset($this->conf['php_ini_cont']);
		foreach ($this->conf['php_ini_cont'] as $num => $line){
			if(preg_match("#$old#", $line)){
				$this->conf['php_ini_cont'][$num] = preg_replace("#$old#", $new, $line);
				$this->conf['php_ini_modified'] = true;
			}
		}
	}

	// PHP_INI_ADD_EXTENSION
	function php_ini_add_extension($path, $directive, $commented=false)
	{
		if(empty($directive)){
			$directive = "extension";
		}

		$ext = basename($path);
		reset($this->conf['php_ini_cont']);

		$new_line = "$directive=$path". $this->conf['nl'];
		if($commented){
			$new_line = ";$new_line";
		}

		$found = false;
		foreach ($this->conf['php_ini_cont'] as $num => $line){
			$line = preg_replace("/;.*$/", "", $line);

			if(preg_match("/\s*$directive\s*=.*$ext/", $line)){
				$found = true;
				$this->conf['php_ini_cont'][$num] = $new_line;
				break;
			}
		}

		if(!$found){
			array_push($this->conf['php_ini_cont'], $new_line);
		}

		$this->install_log("Added new extension to php.ini: $new_line");
		$this->conf['php_ini_modified'] = true;
	}

	// PHP_INI_ADD_ZEND_EXTENSION
	function php_ini_add_zend_extension($path, $directive="")
	{
		if($directive == ""){
			$directive = $this->conf['zend_ext_directive'];
		}
		$this->php_ini_add_extension($path, $directive);
	}

	// PHP_INI_ADD_ZEND_SECTION
	function php_ini_add_zend_section()
	{
		$found = false;
		foreach ($this->conf['php_ini_cont'] as $num => $line){
			$line =  preg_replace("/;.*$/", "", $line);
			if(preg_match("/\s*\[Zend\]\s*/", $line)){
				$found = true;
			}
		}
		if(!$found){
			array_push($this->conf['php_ini_cont'], str_repeat($this->conf['nl'], 2)."[Zend]".$this->conf['nl']);
		}else{
			array_push($this->conf['php_ini_cont'], $this->conf['nl']);
		}
	}

	// PHP_INI_OPEN
	function php_ini_open()
	{
		$this->install_log("Trying to open PHP.ini (".$this->conf['php_ini_path'].")");
		if(empty($this->conf['php_ini_cont'])){
			if($this->my_file_exists($this->conf['php_ini_path'])){
				$this->conf['php_ini_cont'] = file($this->conf['php_ini_path']);
			}
			else {
				$this->install_log("Opened new (empty) PHP.ini");
				$this->conf['php_ini_cont'] = array();
			}
			$this->conf['php_ini_modified'] = false;
			$this->conf['php_ini_opened'] = true;
		}
		else {
			$this->install_log("PHP.ini is already opened");
		}
	}

	// PHP_INI_CONT_CHECK
	function php_ini_cont_check()
	{
		if(!$this->conf['php_ini_opened']){
			$this->zend_bug("php.ini is not opened while trying to close it");
		}
	}

	// PHP_INI_CLOSE
	function php_ini_close()
	{
		$this->php_ini_cont_check();

		if($this->conf['php_ini_modified']){
			$this->my_fwrite($this->conf['php_ini_path'], "w", join("", $this->conf['php_ini_cont']));
		}
		unset($this->conf['php_ini_cont']);
		$this->conf['php_ini_opened'] = false;
	}

	// PHP_INI_BACKUP
	function php_ini_backup()
	{
		$php_ini_backup = $this->my_backup($this->conf['php_ini_path'], true);
		if(isset($php_ini_backup)) {
			$this->conf['php_ini_backup_msg'] = "The original php.ini was backed up to $php_ini_backup";
		}
	}

	// PHP_INI_RELOCATE
	function php_ini_relocate()
	{
		$new_path = $this->conf['prefix']."/etc";
		$new_ini_path = $new_path."/php.ini";
		$this->my_mkdir($new_path);

		$old_ini_path = $this->strip_symlink($this->conf['php_ini_path']);
		$old_path = $this->my_dirname($old_ini_path);

		$msg = "The following configuration changes have been made:\n\n";

		if($old_ini_path != $new_ini_path){

			$this->install_log("Relocating php.ini from: $old_ini_path to: $new_ini_path");

			$this->my_copy($old_ini_path, $new_ini_path);
			$this->my_unlink($this->conf['php_ini_path']);
			$this->my_symlink($new_ini_path, $this->conf['php_ini_path']);

			$msg.= "- The php.ini file has been relocated from $old_path to $new_path.\n\n".
						 "- A symbolic link for the php.ini file has been created in $old_path.\n";

		}
		if(!empty($this->conf['php_ini_backup_msg'])) {
			$msg .= "\n- ".$this->conf['php_ini_backup_msg']."\n";
		}
		
		if(!preg_match("/^\s*$/s", $msg)) {
			$this->msgbox($msg);
		}
	}

	// ------------------
	// INSTALL FUNCTIONS
	// ==================

	// GET_COMPONENT_VERSION
	function get_component_version($name)
	{
		$name .= ".version";
		if(!isset($this->conf['build_info'][$name])) {
			$this->zend_bug("Cannot find version of $name in the package.");
		}

		return $this->conf['build_info'][$name];
	}

	// ADD_PACKAGE_INFO
	function add_package_info()
	{
		$pack_ini = $this->make_path($this->conf['prefix'], "etc", "pack.ini");
		$pack_conf = @parse_ini_file($pack_ini);
		$pack_conf[strtolower(str_replace(" ", "_", $this->conf['product'])).".version"] = $this->get_component_version("package");
		$pack_output = "";
		foreach ($pack_conf as $name => $val) {
			$pack_output .= "$name=$val\n";
		}
		$this->my_touch($pack_ini);
		$this->str2file($pack_output, $pack_ini);
	}

	// ADD_COMPONENT
	function add_component($name, $destination, $make_backup)
	{
		$this->conf['components'][$name] = array($destination, $make_backup);
	}

	// SET_COMPONENTS
	function set_components($new_components)
	{
		$this->conf['components'] = $new_components;
	}

	// REMOVE_COMPONENT
	function remove_component($name)
	{
		unset($this->conf['components'][$name]);
	}

	// SET_VAR_COMPONENT
	function set_var_component($pattern, $value) /* substitute undefined variables
																									in components destinations */
	{
		foreach($this->conf['components'] as $name => $data){
			$this->conf['components'][$name][0] = preg_replace("#$pattern#", $value, $data[0]);
		}
	}

	// START_INSTALL
	function start_install($stop_webserver=true)
	{
		ini_set("max_execution_time", 100000000); /* give reasonable execution time */

		if(false && $this->is_using_apache()) {
			# If there are old copies of files in the installation prefix
			# try to stop the Web server before overwritting them
			foreach (array_keys($this->conf['components']) as $file) {
				if(preg_match("/\.so$/", $file)) {
					if($this->find_file ($this->conf['prefix'], basename($file)) > 0) {
						$msg = 
						"The installation has found installation of Zend product in ".
						$this->conf['prefix']."\n\n".
						"Its recommended to stop your Web server before proceeding with copying\n".
						"new files onto your system. Your Web server will be started automatically\n".
						"by the installation after it finishes to copy files.\n\n".
						"Do you want to stop your Web server now?";

						if($this->yesnobox($msg)) {
							$webserver_stopped = $this->apache_stop ();
						}
						break;
					}
				}
			}
		}

		$percent = 0;
		$size_done = 0;

		$backup_dirs = array(); // array of the directories to backup
		$total_size = $this->gather_components_info($backup_dirs);
		if($total_size == 0) { $total_size = 100; }

		$this->dialog_calc_max_size();
		$gb = new Gaugebox(
				empty($this->conf['dialog']) ? null: $this->conf['dialog'],
				isset($this->conf['max_x'])  ? $this->conf['max_x'] : null,
				isset($this->conf['max_y'])  ? $this->conf['max_y'] : null, "", 0, $this->conf['title']);

		foreach($backup_dirs as $dir){
			$this->my_backup($dir);
		}

		foreach($this->conf['components'] as $what => $data){

			$where = $data[0];	      // path where to install
			$make_backup = $data[1];	// either make a backup or not
			$comp_name = $data[2];		// name of the component
			$file_size = $data[3];		// file size of the component

			$size_done += $file_size;
			$percent = (int)(($size_done/$total_size)*100);

			$gb->update("Installing $comp_name", $percent);

			if($make_backup){
				$this->my_backup("$where/".basename($comp_name));
			}

			if(!$this->my_dir_exists($where)){ /* create the destination directory */
				$this->my_mkdir_long($where);
			}
			$this->my_copy($what, $where); /* perform copy */
		}
		$gb->close();

		if(isset($webserver_stopped)) {
			$this->apache_start();
		}
	}

	// FREEBSD_LIBC_PATCH
	function freebsd_libc_patch()
	{
		if(!$this->my_file_exists("/usr/lib/libc.so.3")){ /* lmutil needs this dso */
			if($this->my_file_exists("/usr/lib/libc.so.4")){
				$this->my_symlink("/usr/lib/libc.so.4", "/usr/lib/libc.so.3");
			}
			else if($this->my_file_exists("/usr/lib/libc.so.5")){
				$this->my_symlink("/usr/lib/libc.so.5", "/usr/lib/libc.so.3");
			}
			else if($this->my_file_exists("/usr/lib/libc.so")){
				$this->my_symlink("/usr/lib/libc.so", "/usr/lib/libc.so.3");
			}
		}
	}

	// GATHER_COMPONENTS_INFO
	function gather_components_info(&$backup_dirs)
	{
		$this->install_log("Gathering components info ...");
		$total_size = 0;

		foreach($this->conf['components'] as $what => $data){

			// LMUTIL SECTION:
			if(basename($what) == "lmutil"){
				if($this->conf['uname']['sysname'] == "Darwin"){ /* don't install lmutil in Darwin */
					return;
				}
				if($this->conf['uname']['sysname'] == "FreeBSD"){ /* make needed symlinks to libraries */
					$this->freebsd_libc_patch();
				}
			}

			$where = $data[0];
			$backup = $data[1];
			$comp_found = false;

			foreach ($this->conf['comp_dirs_arr'] as $dir){

				$comp = $this->nice_path("$dir/$what");

				if($this->my_file_exists($comp)){

					$file_size = filesize($comp);
					$total_size += $file_size;
					$new_comp_arr[$comp] = array($where, $backup, $what, $file_size);
					$comp_found = true;
					break;
				}
				elseif($this->my_dir_exists($comp)){

					$this->pushd($comp);
					$this->get_file_tree(".", $files_arr);

					if($backup){
						array_push($backup_dirs, $where);
					}

					foreach($files_arr as $file => $file_size){
						$total_size += $file_size;
						$dir_dest = $this->nice_path("$where/".$this->my_dirname($file));
						$new_comp_arr[$this->nice_path("$comp/$file")] = array($dir_dest, false, $what, $file_size);
					}

					$this->popd();
					$comp_found = true;
					break;
				}
			}

			if(!$comp_found) {
				$this->zend_bug("Installation component: '$what' was not found.");
			}
		}

		$this->conf['components'] = $new_comp_arr;
		return $total_size;
	}


	// ADD_GROUP
	function add_group($group, $gid)
	{
		$this->install_log("Adding the group: $group to the computer ...");

		if($this->conf['uname']['sysname'] == "Darwin"){
			$groupfile_cont = $this->cmd2str("nicl . -list /groups 2>&1"); // get group list

		}else{
			$groupfile_cont = $this->file2str($this->conf['groupfile']);
		}

		if(preg_match("/$group/m", $groupfile_cont)){
			return;
		}

		if($this->conf['uname']['sysname'] == "Darwin"){
			while(preg_match("/^\s*$gid.*$/m", $groupfile_cont)){
				$gid++;
			}
		}else{
			while(preg_match("/^[^\:]*\:[^\:]*\:$gid:?$/m", $groupfile_cont)){
				$gid++;
			}
		}

		if($this->conf['uname']['sysname'] == "Darwin"){
			$r = $this->run_command ("nicl . -create /groups/$group passwd \"*\"");
			$r=$r+$this->run_command ("nicl . -append /groups/$group gid $gid");

			if($r != 0){
				$this->warn("Couldn't add a group $group with gid $gid (error: $r). Please add a group $group manually.");
			}
		}else{
			$this->my_fwrite($this->conf['groupfile'], "a", "$group:x:$gid:\n");
		}

	}

	// SET_PERMISSIONS
	function set_permissions($file, $mode, $owner="", $group="")
	{
		$file = $this->make_full_path($file, $this->conf['prefix']);

		$this->install_log("Setting permissions on $file: ".
			"mode=$mode, owner=$owner, group=$group");

		if(!$this->my_file_exists($file) && !$this->my_dir_exists($file)){
			$this->on_error("File: $file does not exist.");
		}

		if(!empty($owner)){
			if(!chown($file, $owner)){
				$this->on_error("Cannot set ownership for: $file");
			}
		}
		if(!empty($group)){
			if($group == "zendtech"){
				$this->add_group($group, 200);
			}
			if(!@chgrp($file, $group)){
				$this->warn("Cannot set group $group for file $file");
			}
		}
		if(!empty($mode)){
			if(!chmod($file, octdec($mode))){
				$this->on_error("Cannot set permissions for: $file");
			}
		}
	}
										

	// ADD_PERMISSIONS
	function add_permissions($file, $mode, $owner="", $group="")
	{
		array_push($this->conf['file_perms'], array($file, $mode, $owner, $group));
	}

	// SET_FILE_PERMISSIONS
	function set_file_permissions()
	{
		foreach ($this->conf['file_perms'] as $perms){
			$this->set_permissions ($perms[0], $perms[1], $perms[2], $perms[3]);
		}
	}

	// CRONTAB_ERROR
	function crontab_error($cron_err)
	{
		$answer = $this->yesnobox(
				"WARNING!\n\n$cron_err\n\nWould you like to continue setup anyway?"
				, "NO"
		);
		if(!$answer) {
			$this->default_abort();
		}
	}

	// ADD_TO_CRONTAB
	function add_to_crontab(
		$cron_msg, $user, $line, $cron_warn, $cron_err, $ident = "Zend crontab entry")
	{
		$this->install_log("Trying to add the line: $line to the $user's crontab file.");

		$cron_deny_files = array(
				"/usr/lib/cron/cron.deny",   /* IRIX 5.3  */
				"/etc/cron.d/cron.deny",     /* SunOS 5.X */
				"/var/spool/cron/cron.deny", /* SunOS 4.X */
				"/etc/cron.deny",            /* Linux */
				"/var/adm/cron/cron.deny"    /* HP-UX */
				);

		foreach ($cron_deny_files as $file) {
			if($this->my_file_exists($file)) {
				if(preg_match("/^$user$/m", $this->file2str($file))) {
					/*
					$this->on_error(
							"The installation has detected that user $user is restricted from using\n".
							"the crontab feature on your server. This means that the automated\n".
							"cleaning of old cached data will not work properly. To fix this issue,\n".
							"please remove '$user' from the file '$file', and rerun the installation.\n".
							"If you wish to install the Zend Performance Suite without automated cleaning\n".
							"(NOT recommended), rerun the installation with the --no-crontab switch."
							);
					*/
					$this->crontab_error($cron_err);
					return;
				}
			}
		}

		$crontab_opt = $this->cmd2str("crontab -H 2>&1"); // get crontab options (guess type)
		if(preg_match("@-u\s+user@", $crontab_opt)){
			$crontab_type = 2;
		}
		else if(preg_match("@user@", $crontab_opt)){
			$crontab_type = 1;
		}

		if(isset($crontab_type)){

			if($crontab_type == 1){
				$crontab_cont = $this->cmd2str("crontab -l $user 2>/dev/null");
			}
			else{
				$crontab_cont = $this->cmd2str("EDITOR=cat VISUAL=cat crontab -u $user -e 2>/dev/null");
			}

			$error = "Failed to execute command: ";

			// remove the old settings
			$crontab_cont = preg_replace("@\n?# $ident:.*# End of $ident\n?@s",
					"", $crontab_cont);

			// add the new line
			$crontab_cont .= "\n# $ident:\n$line\n# End of $ident\n";

			while(true)
			{
				if(!$this->yesnobox(
					"$cron_msg\n\nWe are going to add scheduled execution to the $user's crontab file:\n\n".
					wordwrap($line, $this->conf['wrap_size'])."\n\nDo you accept?"))
				{
					if($this->yesnobox($cron_warn)) {
						$this->default_abort();
					}
					continue;
				}
				break;
			}

			$fp = $this->my_fopen($this->conf['tmp_dir']."/crontab", "w");
			fwrite($fp, $crontab_cont);
			fclose($fp);

			if($crontab_type == 1){ // First type of crontab
				putenv("EDITOR=cat ". $this->conf['tmp_dir'] ."/crontab >");
				putenv("VISUAL=cat ". $this->conf['tmp_dir'] ."/crontab >");

				if($this->run_command("crontab -e $user")==0){
					unset($error);
				}
				else { $error .= "crontab -e $user"; }
			}
			else { // Second type
				if($this->run_command("crontab -u $user ". $this->conf['tmp_dir'] ."/crontab")==0){
					unset($error);
				}
				else { $error .= "crontab -u $user ". $this->conf['tmp_dir'] ."/crontab"; }
			}
		}
		else {
			$error = "Failed to execute crontab.";
		}

		if(isset($error)){
			//$this->warn("Couldn't add a crontab scheduled execution.\n$error");
			$this->crontab_error($cron_err);
		}
	}


	// --------------------------
	// LICENSE FUNCTIONS
	// ==========================

	// SHOW_LICENSE
	function show_license($file, $message="\n")
	{
		while(true) {
			$this->show_file($file);
			if(!$this->yesnobox($message."Do you accept the terms of the license?", "YES")) {
				if($this->yesnobox("In order to install the ".$this->conf['product']." you must accept this agreement!\n".
							"Do you really want to cancel the installation?", "NO")) {
					$this->on_abort("", true);
				}
			}
			else {
				break;
			}
		}
	}

	// LICENSE_AGMNT_BOX
	function license_agmnt_box()
	{
		$this->install_log("Showing license: ".$this->conf['license_agr_file']);

		$license = "IMPORTANT:\n".
			"BY SELECTING THE 'YES' OPTION BELOW, DOWNLOADING, INSTALLING, OR\n".
			"OTHERWISE USING THIS SOFTWARE, YOU ACKNOWLEDGE THAT YOU HAVE READ THE\n".
			"LICENSE AGREEMENT, AND THAT YOU AGREE TO BE BOUND BY ITS TERMS AND\n".
			"CONDITIONS.\n".
			"IF YOU DO NOT AGREE TO ALL OF THE TERMS AND CONDITIONS OF SUCH AGREEMENT,\n".
			"YOU ARE NOT AN AUTHORIZED USER OF THE SOFTWARE AND IT IS YOUR\n".
			"RESPONSIBILITY TO EXIT THIS DOWNLOADING/INSTALLATION PROCESS WITHOUT\n".
			"DOWNLOADING OR INSTALLING THE SOFTWARE BY SELECTING THE 'NO' OPTION BELOW,\n".
			"AND TO DELETE THE SOFTWARE FROM YOUR COMPUTER.\n\n\n";

		$this->show_license($this->conf['license_agr_file'], $license);
	}

	// VERIFY_KEY
	function verify_key($user, $key)
	{
		$rc = $this->run_command("./verifykey \"$user\" $key");
		if($rc == 0){
			$this->zend_bug("Wrong parameters passed to verifykey.");
		}
		if($rc == 127){
			$this->on_error("Could not execute verifykey.");
		}

		return $rc;
	}

	// DLG_VERIFY_LICENSE_KEY
	function dlg_verify_license_key(&$user, &$key)
	{
		if(isset($this->conf['license_user'])) {
			$user = $this->conf['license_user'];
		}
		if(isset($this->conf['license_key'])) {
			$key = $this->conf['license_key'];
		}

		$rc = 0;
		do{
			switch($rc){
				case    0:  $error = "";
										break;
				case  255:  $error = "The name and/or key you have entered are incorrect !\n\n";
										break;
				case  254:  $error = "The key you have entered has expired !\n\n";
										break;
				default  :  $error = "Unrecognized return value from verifykey !\n\n";
			}
			$user = $this->inputbox(
					$error."Please enter the Registration Name you received for the ".$this->conf['product'], $user);

			while(true){
				$key = $this->inputbox("Please enter the License Key you received for the ".$this->conf['product'], $key);
				if(!preg_match("/^[A-Za-z0-9]+$/", $key)){
					$this->msgbox("The key you entered contains invalid characters.  Please try again!");
				}
				else{
					$key = preg_replace("/^LK/i", "", trim($key));
					if(!preg_match("/^[A-Za-z0-9]+$/", $key)){
						$this->msgbox("Invalid key.  Please try again!");
					}else{
						break;
					}
				}
			}
		}
		while(($rc = $this->verify_key($user, $key)) != 1);
	}

	// GEN_LICENSE_URL
	function gen_license_url($args)
	{
		$urlstr = "";
		foreach($args as $key => $value){
			$urlstr .= "$key=".urlencode($value)."&";
		}
		$urlstr = rtrim($urlstr, "&");

		return "https://".$this->conf['license_host']."/".$this->conf['license_script']."?$urlstr";
	}

	// GEN_LICENSE
	function gen_license($args, $file)
	{
		error_reporting(0);
		set_error_handler("error_handler");

		$err = 0;
		$this->my_unlink($this->conf['lic_response_file']);

		$args["pass"] = md5($args["pass"]);
		$url = $this->gen_license_url($args);
		$this->install_log("Generated URL: $url");

		if(!extension_loaded("curl")){
			$this->zend_bug("Your PHP must be loaded with cURL extension.");
		}
		/* start cURL session (Secure HTTP) */
		$ch = curl_init();
		curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, 0);
		curl_setopt($ch, CURLOPT_TIMEOUT, 300);
		curl_setopt($ch, CURLOPT_URL, $url);
		curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);
		$buf = curl_exec($ch);

		$curl_errno = curl_errno($ch);

		if(!$buf || $curl_errno != CURLE_OK){

			$curl_error = "The installation failed to connect to the Zend License Server.\n";
			switch($curl_errno) {
				case CURLE_COULDNT_CONNECT:
				case CURLE_OPERATION_TIMEOUTED:
					$curl_error .= "This may be due to a local or remote connection failure.\n";
					break;
				case CURLE_COULDNT_RESOLVE_HOST:
					$curl_error .= "The host ". $this->conf['license_host'] ." couldn't be resolved.\n";
					break;
				default:
					$curl_error = curl_error($ch);
			}
			$cur_error .= "Please check that you are connected to the Internet or try again\nin a few minutes.";

			if(!$this->my_fwrite($this->conf['lic_response_file'], "w", "$curl_error\n")){
				return -2;
			}
			return (-1);
		}
		curl_close($ch);
		/* end of cURL session */

		if(empty($buf)){
			if(!$this->my_fwrite($this->conf['lic_response_file'], "w", "Server returned empty buffer\n")){
				return -2;
			}
			return (-1);
		}

		$lines = preg_split("/\n/", $buf);

		/* create response file */
		if(!$this->my_fwrite($this->conf['lic_response_file'], "w", "$lines[0]\n$lines[1]\n")){
			return -2;
		}

		/* create license file */
		$buf = "";
		for($i=2; !empty($lines[$i]); $i++){
			$buf .= "$lines[$i]\n";
		}

		/* create response file */
		if(!$this->my_fwrite($file, "w", $buf)){
			return -2;
		}

		restore_error_handler();
		error_reporting(E_ALL);
		return 0;
	}

	// ERROR_HANDLER
	function error_handler($errno, $errstr, $errfile, $errline)
	{
		$this->my_fwrite($this->conf['lic_response_file'], "a", "LOCAL ERROR: $errstr in $errfile line: $errline\n");
	}

	// GET_LMUTIL_HOST_ID
	function get_lmutil_host_id()
	{
		$line = $this->cmd2str("../lmutil lmhostid");
		if(!preg_match("/^.+\"(.+)\"/m", $line, $hostid)){
			$this->on_error("Can't determine host ID.");
		}

		return $hostid[1];
	}

	// GET_ZENDID_HOST_ID
	function get_zendid_host_id()
	{
		$array=split("\n",$this->cmd2str("../data/zendid allid"));
		$comma_separated_array = implode(",", $array);
		return $comma_separated_array;
	}

	// GET_HOST_ID
	function get_host_id()
	{
		return $this->get_zendid_host_id();
	}

	// GET_RESPONSE
	function get_response(&$code, &$status)
	{
		if(!($fp = fopen($this->conf['lic_response_file'], "r"))){
			$this->on_error("Couldn't open ".$this->conf['lic_response_file']."for reading");
		}

		$code = preg_replace("/(^.*)[\r\n]/", "\\1", fgets($fp, 1024));
		$status = preg_replace("/(^.*)[\r\n]/", "\n\\1\n", fgets($fp, 1024));
		$status = str_replace("^", "\n", $status);

		if(preg_match("#\s*<HTML>#i", $status)){
			$status = "";
		}

		fclose($fp);
	}

	// GET_LICENSE
	function get_license( &$passwd, &$uid, $license_file = "")
	{
		if($license_file == ""){
			$license_file = $this->conf['license_file'];
		}

		// check for license in package
		if($this->my_file_exists("../data/$license_file")){
			return 0;
		}

		$args = array();
		static $save_args;

		if(empty($this->conf['release'])){
			$this->zend_bug("Undefined release name");
		}
		if(empty($this->conf['product_id'])){
			$this->zend_bug("Undefined product ID");
		}

		if($this->conf['release'] == "release"){
			$release = "0";
		}
		elseif(preg_match("/^beta(.)$/", $this->conf['release'], $arr)){
			$release = $arr[1];
		}
		else{
			$release = "99"; // invalid build_id
		}

		$build_id = $this->conf['version']."-".$release;
		$args["version"] = $build_id;

		if(extension_loaded("curl") && $this->conf['interactive']){
			$m1 = "Download a license file from ".$this->conf['license_host'];
			$m2 = "Search for a license file on my disk";
			$menu = array(1 => $m1, 2 => $m2);

			$rc = $this->menubox(wordwrap(
						"In order to proceed, Setup needs a valid license file. ".
						"The license file is called: \"$license_file\". ".
						"If you don't have a valid license file, select \"$m1\". ".
						"Otherwise select \"$m2\". ".
						"The license file can also be downloaded manually from: ".
						"http://www.zend.com/store/pickup.php.", $this->conf['wrap_size']),
					$menu
					);
		}
		else if(!$this->conf['interactive']) {
			$rc = 1;
		}
		else{
			$rc = 2;
		}

		if($rc == 2){
			$this->find_license_on_disk($license_file);
		}
		elseif($rc == 1){
			if(defined('SERIAL_NUMBER')){
				if($this->dlg_existing_user_ask()){
					$this->dlg_username_password($args);
				}
				else{
					if(isset($save_args)) {
						$args = $save_args;
					}
					$this->register_user($args);

					$save_args = $args;
				}

				if(!defined('TRIAL')) {
					$trial_commerc = $this->dlg_trial_commercial($args);
					if($trial_commerc == "c"){
						$this->dlg_ask_serial_number($args);
					}
					else{
						$args["serial"]="";
					}
				}
				else{
					$args["serial"]="";
				}
			}
			else{
				$this->dlg_username_password($args, "(If you don't have a Zend username,\n".
						"please register at www.zend.com/add_user.php)\n\n");
			}
			unset ($args["passc"]);
			$ret_value=$this->download_license($license_file, $args);

			if($ret_value != -1) {
				$passwd=$args["pass"];
				$uid=$args["uid"];
			}
			else {
				$passwd = "";
				$uid = "";
			}
			if(defined('SERIAL_NUMBER') && ($ret_value!=-1))
			{
				if(!defined('TRIAL')){ // trial

					$this->msgbox("Successfully downloaded license file.\n".
							"In case you may wish to download your license information in the future,\n".
							"you may do so by accessing http://www.zend.com/store/pickup.php.");

				} else {
					$this->msgbox("Successfully downloaded a ".trim($ret_value)."-day trial license.");
				}
				return 0;
			}
			return $ret_value;
		}
		return 0;
	}

	// REGISTER_USER
	function register_user(&$args)
	{
		do
		{
			$this->forms->reg_page1($args);
		}
		while(!$this->forms->reg_page2($args));
	}

	// REGISTER_PHPEE_SERVER
	function register_phpee_server(&$args)
	{
		$this->forms->register_phpee_server($args);
	}

	// DLG_ASK_SERIAL_NUMBER
	function dlg_ask_serial_number(&$args)
	{
		if(isset($this->conf['serial_number'])) {
			$serial = $this->conf['serial_number'];
		}
		else {
			do
			{
				$serial = $this->inputbox("Please enter your serial number. The serial number can be found at\n".
						"http://www.zend.com/store/pickup.php or in the purchase confirmation E-mail");

				if(strlen($serial) < 8){
					$this->msgbox("Serial number must contain 8 or more characters");
				}
				elseif(!preg_match ("/^SN/i",$serial)){
					$this->msgbox("Serial number should start with SN");
				}
				else break;
			}
			while(true);
		}

		$args["serial"] = $serial;
	}

	// DLG_TRIAL_COMMERCIAL
	function dlg_trial_commercial(&$args)
	{
		if(!$this->conf['interactive']) {
			return "c";
		}

		if(defined('TRIAL')){ /* return immidiately, if the product is for trial usage */
			return "t";
		}
		$m = array(
				"t" => "Get Trial License",
				"c" => "Get Commercial License"
				);
		$r = $this->menubox("Is the product for trial or commercial use?", $m, false, "", "c");
		return $r;
	}

	// DLG_EXISTING_USER_ASK
	function dlg_existing_user_ask()
	{
		if(!isset($this->conf['existing_zend_com_account'])) {
			$this->conf['existing_zend_com_account'] = $this->yesnobox("Do you have a Zend.com account?\n".
				"You will be asked for your Zend.com username\nand password during installation.");
		}
		return $this->conf['existing_zend_com_account'];
	}

	// DLG_USERNAME_PASSWORD
	function dlg_username_password(&$args, $msg="")
	{
		$username_def = "";
		if(isset($this->conf['zend_username'])) {
			$username_def = $this->conf['zend_username'];
		}

		$password_def = "";
		if(isset($this->conf['zend_password'])) {
			$password_def = $this->conf['zend_password'];
		}

		$this->conf['zend_username'] = $this->inputbox($msg."Please enter your Zend.com username", $username_def);
		$this->conf['zend_password'] = $this->password_get("Please enter your Zend.com password: ");

		if(empty($this->conf['dialog'])){
			print("\n");
		}

		$args["uid"] = $this->conf['zend_username'];
		$args["pass"] = $this->conf['zend_password'];
	}

	// DOWNLOAD_LICENSE
	function download_license($license_file, $args)
	{
		$args["pid"] = $this->conf['product_id'];
		$args["host"] = $this->get_host_id();
		$args["ZLM"] = "1";

		$this->progress_bar_start(300,
				"Downloading license. Please wait...\n(this may take a few minutes)",
				"Downloading ..."
				);
		$r = $this->gen_license ($args, $license_file);
		$this->progress_bar_stop();

		if($r != 0){
			switch($r){
				case -1:
					$msg = $this->file2str($this->conf['lic_response_file']);
					break;

				case -2:
					$this->on_error("Could not write response file.");
					break;

				default:
					$this->on_error("Impossibe return value from gen_license(): $r");
			}
			return $this->license_errmsg($msg);
		}

		$this->get_response(&$code, &$status);

		if($code != "OK") {
			$this->my_unlink($license_file);
			return $this->license_errmsg($status);
		}

		// license file needs to end up in 'data' for install_license()
		$this->my_copy($license_file, "../data/$license_file");

		if(!defined('SERIAL_NUMBER')) {
			if(isset($status) && $status != -1 && preg_match("/^\s*\d+$/", $status)) {
				$this->msgbox("Successfully downloaded a ".trim($status)."-day trial license!");
			}
			else {
				$this->msgbox("Successfully downloaded license file!");
			}
		}
		return $status;

	}

	// FIND_LICENSE_ON_DISK
	function find_license_on_disk($license_file="", $product="", $license_def_dir="")
	{
		if(empty($license_file)){
			$license_file = $this->conf['license_file'];
		}

		// check for license in package
		if($this->my_file_exists("../data/$license_file")){
			return 0;
		}

		// look for a license in current directory
		if($this->my_file_exists("./$license_file")){
			$this->my_move($license_file, "../data/$license_file");
			return 0;
		}

		if(empty($product)){
			$product = $this->conf['product'];
		}

		$license_dir="";
		while(true){
			$license_dir = $this->choose_dir(
					"Specify a path to a $product license file\n(it's called: $license_file)", $license_def_dir);

			if($this->my_file_exists($license_dir)) {
				if(basename($license_dir) != $license_file) {
					$this->msgbox("License file must be called: $license_file !");
				}
				else {
					$license_dir = dirname($license_dir);
					break;
				}
			}
			else if(!$this->my_file_exists($license_dir."/$license_file")){
				$this->msgbox("License file:\n$license_dir/$license_file\n".
						"does not exist. Please try again.");
			}
			else break;
		}

		$this->my_copy($license_dir."/".$license_file, "../data/$license_file");

		return $license_dir;
	}

	// LICENSE_ERRMSG
	function license_errmsg($msg)
	{
		$this->msgbox(
				"There was a problem downloading the license file:\n\n$msg\n" .
				"If you still can't download a license file, you can download it from\n".
				"http://www.zend.com/store/pickup.php, or email the Zend sales team at\n".
				"sales@zend.com."
				);
		return -1;
	}

	// INSTALL_LICENSE
	function install_license($prefix="", $license="")
	{
		if(empty($prefix)){
			$prefix = $this->conf['prefix'];
		}

		if(empty($license)){
			$license = $this->conf['license_file'];
		}

		if(!$this->my_file_exists("../data/$license")){
			$this->on_error("License file: $license does not exist");
		}
		$this->my_mkdir_long($prefix);

		$this->infobox("Installing license $license ...");
		$this->my_copy("../data/$license", $prefix);

		$this->conf['license_installed'] = true;
	}


	// ------------------
	// BUGS FUNCTIONS
	// ==================

	// ZEND_BUG
	function zend_bug($msg="")
	{
		$this->msgbox("$msg\n\nPlease send a bug report to: <".$this->conf['packaging_dot_zend'].">");
		if($this->yesnobox("Send it now?")){
			$this->zend_send_bug($msg);
		}

		$this->on_error($msg);
	}

	// ZEND_SEND_BUG
	function zend_send_bug($msg="")
	{
		$from="";
		$from = $this->inputbox("Please provide an e-mail address for Zend response (Optional)", "", true);

		$this->install_log("Sending bug report to: <".$this->conf['packaging_dot_zend']."> from $from");
		$this->infobox("Sending mail ...");

		$logfile = $this->conf['logfile_name'];
		/* Uncomment this to enable sending the entire var_dump($this->conf) :
			 $logfile = $this->conf['tmp_dir']."/install.log";
			 $this->str2file($this->file2str($this->conf['logfile_name']).
			 "\nVAR DUMP:\n".var_export($this->conf, true), $logfile);
		 */

		if($this->send_mail($this->conf['packaging_dot_zend'], "Zend installation bug",
					"ERROR: $msg\nSee attached installation logfile for more information.\n", $logfile,
					$from, "text/plain")){

			$this->msgbox("Bug report sent successfully.");
		}
		else{
			$this->msgbox("Cannot send bug report!");
		}
	}

	// SEND_MAIL
	function send_mail($to, $subject, $message, $file_name, $from, $mime_type="application/octet-stream")
	{
		// Generate a boundary string
		$semi_rand = md5(time());
		$mime_boundary = "Zend-$semi_rand";

		$smtp_headers = "";
		/*
			 $smtp_headers = "From: user@domain\n".
			 "Reply-To: user@domain\n".
			 "X-Mailer: PHP4\n".
			 "X-Sender: user@domain\n";
		 */

		$text_body = "--$mime_boundary\n".
			"Content-Type: text/plain; charset=\"us-ascii\"\n\n".
			"$message\n";

		// Open file to attach
		$fp = $this->my_fopen($file_name, "rb");
		$data = fread($fp, filesize($file_name));
		fclose($fp);

		// Prepare an encoded message
		$text_encoded = "--$mime_boundary\n".
			"Content-type: $mime_type; name=\"".basename($file_name)."\";\n".
			"Content-Transfer-Encoding: base64\n".
			"Content-disposition: attachment; filename=\"".basename($file_name)."\"\n\n".
			$this->my_chunk_split(base64_encode($data))."\n".
			"--$mime_boundary--\n";

		$mime_headers = "MIME-version: 1.0\n".
			"Content-type: multipart/mixed; ".
			"boundary=\"$mime_boundary\"\n".
			"Content-transfer-encoding: 7BIT\n".
			"X-attachments: $file_name;\n\n";


		// Send the message
		$headers = $smtp_headers . $mime_headers;
		$message = $text_body . $text_encoded;
		if($from !=""){
			return @mail($to, $subject, $message, "From: $from \r\n".$headers);
		}else{
			return @mail($to, $subject, $message, $headers);
		}
	}

	// MY_CHUNK_SPLIT
	function my_chunk_split($str)
	{
		$stmp = $str;
		$len = strlen($stmp);
		$out = "";
		while ($len > 0) {
			if ($len >= 76) {
				$out = $out . substr($stmp, 0, 76) . "\r\n";
				$stmp = substr($stmp, 76);
				$len = $len - 76;
			}
			else {
				$out = $out . $stmp . "\r\n";
				$stmp = ""; $len = 0;
			}
		}
		return $out;
	}

	// SEARCH_CMD_IN_PATH
	function search_cmd_in_path($cmd)
	{
		$cmd_path = null;

		$path_arr = explode(":", getenv("PATH"));
		if(is_array($path_arr)) {
			foreach ($path_arr as $path) {
				$tmp_path = $this->make_path ($path, $cmd);
				if($this->my_file_exists ($tmp_path)) {
					$cmd_path = $tmp_path;
					break;
				}
			}
		}
		return $cmd_path;
	}

	// DETECT_IP_ADDRESS
	function detect_ip_address()
	{
		$ext_ip = null;
		$def_route_line = shell_exec("netstat -r");

		if(preg_match("/^\s*default.*\s([^\s]+)\s*$/m", $def_route_line, $match)) {
			$ext_int = $match[1];
		}
		if(isset($ext_int)) {
			$ifconfig_out = shell_exec("/sbin/ifconfig $ext_int");
			if(preg_match("/addr\:([\d\.]+)/", $ifconfig_out, $match)) {
				$ext_ip = $match[1];
			}
		}
		return $ext_ip;
	}

	// SEPARATE_IP_PORT
	function separate_ip_port($url, &$ip, &$port)
	{
		$arr = explode(":", $url);
		$ip = $arr[0];
		$port = isset($arr[1]) ? $arr[1] : null;
	}

	// KILL_PROCESS
	function kill_process ($name, $signal="KILL")
	{
		$status = 1;

		$ps_res = array();
		@exec ("ps ax 2>/dev/null", &$ps_res);
		foreach ($ps_res as $ps_line) {
			if(preg_match ("/(\d+).*$name/", $ps_line, $match)) {
				$status = $this->run_command ("kill -$signal ".$match[1]);
			}
		}
		if($status != 0) {
			$ps_res = array();
			@exec ("ps -ef 2>/dev/null", &$ps_res);
			foreach ($ps_res as $ps_line) {
				if(preg_match ("/(\d+).*$name/", $ps_line, $match)) {
					$status = $this->run_command ("kill -$signal ".$match[1]);
				}
			}
		}
		if($status != 0) {
			$killall = $this->search_cmd_in_path ("killall");
			if($killall) {
				$status = $this->run_command ("$killall -$signal $name");
			}
		}
		if($status == 0) {
			$this->install_log ("Killed process: $name with signal: $signal");
		}
		else {
			$this->install_log ("Failed to kill process: $name");
		}
	}

	// WRITE_VHOSTS_FILE
	function write_vhosts_file($vhosts_file)
	{
		if($this->my_file_exists($vhosts_file)) {
			$vhosts_arr = @file($vhosts_file);
		}
		else $vhosts_arr = array();

		//create vhosts.ini for configuration vr_hosts :
		//host:ip;document_root

		foreach ($vhosts_arr as $num => $line) {
			if(strncmp($line, "http://", 7)==0 || strncmp($line,"https://", 8)==0) {
				continue;
			}

			$proto = "http";
			if(preg_match("/:(\d+)/", $line, $match)) {
				if($match[1] == "443"){
					$proto = "https";
				}
			}
			$vhosts_arr[$num] = "$proto://$line";
		}

		if(isset($this->conf['apache_vhosts'])) {
			foreach ($this->conf['apache_vhosts'] as $vhost_key => $vhost_docroot_array) {

				$vhost_ip=$vhost_docroot_array['ip'];
				$vhost_port=$vhost_docroot_array['port'];
				$vhost_servername=$vhost_docroot_array['server'];
				$vhost_docroot = $vhost_docroot_array['docroot'];

				if($vhost_port == "443") {
					$proto = "https";
				}
				else $proto = "http";

				if($vhost_port == "80") {
					$vhost_port = "";
				}
				else $vhost_port = ":$vhost_port";

				if (preg_match("/\d+\.\d+\.\d+\.\d+/",$vhost_ip)) {
					$vhost_servername = $vhost_ip;
				}
				array_push($vhosts_arr, "$proto://$vhost_servername$vhost_port;$vhost_docroot\n");
			}
		}

		$this->my_fwrite($vhosts_file, "w", implode("", array_unique($vhosts_arr)));
	}

	// EXEC_STR_THROUGH_WEBSERVER
	function exec_str_through_webserver($str, $ext = "php")
	{
		$test_file = rand().".".$ext;
		$test_file_full = $this->make_path($this->webserver_docroot_guess(), $test_file);

		$this->str2file($str, $test_file_full);

		$this->install_log("Executing: $str trough the webserver");

		$host = $this->apache_get_virtual_host("127.0.0.1");

		$cont = @file_get_contents("http://".$host."/".$test_file);
		$this->my_unlink($test_file_full);

		$this->install_log("Webserver returned: $cont");

		return $cont;
	}

	// CHECK_PHP_IS_LOADED
	function check_php_is_loaded()
	{
		$test_str = "<? print('PHP_IS_LOADED'); ?>";
		return strstr($this->exec_str_through_webserver($test_str), "PHP_IS_LOADED");
	}

	// CHECK_WEBSERVER_IS_LOADED
	function check_webserver_is_loaded()
	{
		$test_str = "WEBSERVER_IS_LOADED";
		return strstr($this->exec_str_through_webserver($test_str, "html"), "WEBSERVER_IS_LOADED");
	}

	// CHECK_LOADED_EXTENSIONS
	function check_loaded_extensions()
	{
		$test_str = "<? print(join('\n', get_loaded_extensions())); ?>";
		$cont = $this->exec_str_through_webserver($test_str);
		return split("\n", $cont);
	}

	// POSTINSTALL_TEST
	function postinstall_test($test_extensions)
	{
		$this->webserver_docroot_guess();

		$this->progress_bar_start(100,
			"Running post installation tests\n(this may take a few seconds)...", "Testing ...");

		if(!$this->check_webserver_is_loaded()) {
			$this->progress_bar_stop();
			$this->msgbox(
				"Setup was unable to communicate with the Web server. It is possible that\n".
		    "the Web server failed to start. Try to start the Web server manually."
			);
		}
		else {
			if(!$this->check_php_is_loaded()) {
				$this->progress_bar_stop();
				$this->msgbox("Setup was unable to detect working PHP on your Web server.");
			}
			else {
				if(@is_array($test_extensions) && @count($test_extensions) > 0) {
					$extensions = $this->check_loaded_extensions();
					
					$post_summary = $this->conf['product']." Post Installation Summary:\n\n";
					foreach ($test_extensions as $ext) {
						$post_summary .= "$ext (".(in_array($ext, $extensions) ? "Success" : "Failure").")\n";
					}
				}
				$this->progress_bar_stop();
			}
		}

		if(isset($post_summary)) {
			$this->msgbox($post_summary);
		}
	}
}

function error_handler($errno, $errstr, $errfile, $errline)
{
	global $INSTALL;
	$INSTALL->error_handler($errno, $errstr, $errfile, $errline);
}

?>
